from dungeonsheets.features.features import Feature, FeatureSelector
from dungeonsheets.features.ranger import Archery, Defense, Dueling, TwoWeaponFighting

# Features added for all PHB classes
# SCAG and XGTE needed


# PHB
class GreatWeaponFighting(Feature):
    """When you roll a 1 or 2 on a damage die for an attack you make with a melee
    weapon that you are wielding with two hands, you can reroll the die and
    must use the new roll, even if the new roll is a 1 or a 2. The weapon must
    have the two-handed or versatile property for you to gain this benefit.
    """

    name = "Fighting Style (Great Weapon Fighting)"
    source = "Fighter"


class Protection(Feature):
    """When a creature you can see attacks a target other than you that is within
    5 feet of you, you can use your reaction to impose disadvantage on the
    attack roll. You must be wielding a shield.
    """

    name = "Fighting Style (Protection)"
    source = "Fighter"


# UA
class BlindingFighting(Feature):
    """Being unable to see a creature doesn’t impose disadvantage on your attack
    rolls against it, provided the creature isn’t hidden from you.
    """

    name = "Fighting Style (Blinding Fighting)"
    source = "Fighter"


class CloseQuartersShooter(Feature):
    """When making a ranged attack while you are within 5 feet of a hostile
    creature, you do not have disadvantage on the attack roll. Your ranged
    attacks ignore half cover and three-quarters cover against targets within
    30 feet of you. You have a +1 bonus to attack rolls on ranged attacks.
    """

    name = "Fighting Style (Close Quarters Shooter)"
    source = "Fighter"


class Interception(Feature):
    """When a creature you can see hits a target that is within 5 feet of you
    with an attack, you can use your reaction to reduce the damage the target
    takes by 1d10 + your proficiency bonus (to a minimum of 0 damage). You must
    be wielding a shield or a simple or martial weapon to use this reaction.
    """

    name = "Fighting Style (Interception)"
    source = "Fighter"


class Mariner(Feature):
    """As long as you not wearing heavy armor or using a shield, you have
    a swimming speed and a climbing speed equal to your normal speed, and you
    gain a +1 bonus to armor class.
    """

    name = "Fighting Style (Mariner)"
    source = "Fighter"


class ThrownWeaponFighting(Feature):
    """You can draw a weapon that has the thrown property as part of the
    attack you make with the weapon.

    In addition, when you hit with a ranged attack using a thrown
    weapon, you gain a +1 bonus to the damage roll.

    """

    name = "Fighting Style (Thrown Weapon Fighting)"
    source = "Fighter"


class TunnelFighter(Feature):
    """As a bonus action, you can enter a defensive stance that lasts until
    the start of your next turn. While in your defensive stance, you can make
    opportunity attacks without using your reaction, and you can use your
    reaction to make a melee attack against a creature that moves more than
    5 feet while within your reach.
    """

    name = "Fighting Style (Tunnel Fighter)"
    source = "Fighter"


class UnarmedFighting(Feature):
    """Your unarmed strikes can deal bludgeoning damage equal to 1d6 + your
    Strength modifier. If you strike with two free hands, the d6 becomes a d8.

    When you successfully start a grapple, you can deal 1d4
    bludgeoning damage to the grappled creature. Until the grapple
    ends, you can also deal this damage to the creature whenever you
    hit it with a melee attack.

    """

    name = "Fighting Style (Unarmed Fighting)"
    source = "Fighter"


class FighterFightingStyle(FeatureSelector):
    """Select a Fighting Style by choosing in feature_choices:

    - archery
    - defense
    - dueling
    - great-weapon fighting
    - protection
    - two-weapon fighting
    - blinding fighting
    - close quarters shooter
    - interception
    - mariner
    - thrown weapon fighting
    - tunnel fighter
    - unarmed fighting

    """

    options = {
        "archery": Archery,
        "defense": Defense,
        "dueling": Dueling,
        "great": GreatWeaponFighting,
        "great-weapon fighting": GreatWeaponFighting,
        "projection": Protection,
        "two-weapon fighting": TwoWeaponFighting,
        "two-weapon": TwoWeaponFighting,
        "dual wield": TwoWeaponFighting,
        "blinding fighting": BlindingFighting,
        "close quarters shooter": CloseQuartersShooter,
        "interception": Interception,
        "mariner": Mariner,
        "thrown weapon fighting": ThrownWeaponFighting,
        "tunnel fighter": TunnelFighter,
        "unarmed fighting": UnarmedFighting,
    }
    name = "Fighting Style (Select One)"
    source = "Fighter"


class SecondWind(Feature):
    """You have a limited well of stamina that you can draw on to protect yourself
    from harm. On your turn, you can use a bonus action to regain hit points
    equal to 1d10 + your fighter level. Once you use this feature, you must
    finish a short or long rest before you can use it again

    """

    name = "Second Wind"
    source = "Fighter"


class ActionSurge(Feature):
    """Starting at 2nd level, you can push yourself beyond your normal limits for a
    moment. On your turn, you can take one additional action on top of your
    regular action and a possible bonus action.

    Once you use this feature, you must finish a short or long rest before you
    can use it again. Starting at 17th level, you can use it twice before a
    rest, but only once on the same turn.

    """

    name = "Action Surge"
    source = "Fighter"


class ExtraAttackFighter(Feature):
    """Beginning at 5th level, you can attack twice, instead of once, whenever you
    take the Attack action on your turn. The number of attacks increases to
    three when you reach 11th level in this class and to four when you reach
    20th level in this class.

    """

    _name = "Extra Attack"
    source = "Fighter"

    @property
    def name(self):
        level = self.owner.Fighter.level
        if level < 11:
            return self._name + " (2x)"
        elif level < 20:
            return self._name + " (3x)"
        else:
            return self._name + " (4x)"


class Indomitable(Feature):
    """Beginning at 9th level, you can reroll a saving throw that you fail. If you
    do so, you must use the new roll, and you can't use this feature again
    until you finish a long rest.

    You can use this feature twice between long rests starting at 13th level
    and three times between long rests starting at 17th level.

    """

    _name = "Indomitable"
    source = "Fighter"

    @property
    def name(self):
        level = self.owner.Fighter.level
        if level < 13:
            return self._name + " (1x/LR)"
        elif level < 17:
            return self._name + " (2x/LR)"
        else:
            return self._name + " (3x/LR)"


# Champion
class ImprovedCritical(Feature):
    """Beginning when you choose this archetype at 3rd level, your weapon attacks
    score a critical hit on a roll of 19 or 20.

    """

    name = "Improved Critical"
    source = "Fighter (Champion)"


class RemarkableAthlete(Feature):
    """Starting at 7th level, you can add half your proficiency bonus (round up)
    to any Strength, Dexterity, or Constitution check you make that doesn't
    already use your proficiency bonus.

    In addition, when you make a running long jump, the distance you can cover
    increases by a number of feet equal to your Strength modifier.

    """

    name = "Remarkable Athlete"
    source = "Fighter (Champion)"


class AdditionalFightingStyle(FeatureSelector):
    """
    Select a Fighting Style by choosing in feature_choices:

    archery 2

    defense 2

    dueling 2

    great-weapon fighting 2

    protection 2

    two-weapon fighting 2
    """

    options = {
        "archery 2": Archery,
        "defense 2": Defense,
        "dueling 2": Dueling,
        "great 2": GreatWeaponFighting,
        "great-weapon fighting 2": GreatWeaponFighting,
        "projection 2": Protection,
        "two-weapon fighting 2": TwoWeaponFighting,
        "two-weapon 2": TwoWeaponFighting,
        "dual wield 2": TwoWeaponFighting,
    }
    name = "Fighting Style (Select One)"
    source = "Fighter (Champion)"


class SuperiorCritical(Feature):
    """Starting at 15th level, your weapon attacks score a critical hit on a roll
    of 18-20 .
    """

    name = "Superior Critical"
    source = "Fighter (Champion)"


class Survivor(Feature):
    """At 18th level, you attain the pinnacle of resilience in battle. At the
    start of each of your turns, you regain hit points equal to 5 + your
    Constitution modifier if you have no more than half of your hit points
    left. You don't gain this benefit if you have 0 hit points.

    """

    name = "Survivor"
    source = "Fighter (Champion)"


# Battle Master
class CombatSuperiority(Feature):
    """When you choose this archetype at 3rd level, you learn maneuvers that are
    fueled by special dice called superiority dice.

    **Maneuvers**: You learn three maneuvers of your choice, which are detailed
    under "Maneuvers" below. Many maneuvers enhance an attack in some way. You
    can use only one maneuver per attack. You learn two additional maneuvers
    of your choice at 7th, 10th, and 15th level. Each time you learn new
    maneuvers, you can also replace one maneuver you know with a different one.

    **Superiority Dice**: You have four superiority dice, which are d8s. A
    superiority die is expended when you use it. You regain all of your
    expended superiority dice when you finish a short or long rest. You gain
    another superiority die at 7th level and one more at 15th level.

    **Saving Throws**: Some of your maneuvers require your target to make a
    saving throw to resist the maneuver's effects. The saving throw DC is
    calculated as follows:

    Maneuver save DC = 8 + your proficiency bonus + your Strength or Dexterity
    modifier (your choice)

    """

    _name = "Combat Superiority"

    @property
    def name(self):
        level = self.owner.Fighter.level
        if level < 10:
            return self._name + " (d8)"
        elif level < 18:
            return self._name + " (d10)"
        else:
            return self._name + " (d12)"


class StudentOfWar(Feature):
    """At 3rd level, you gain proficiency with one type of artisan's tools of your
    choice.

    """

    name = "Student of War"
    source = "Fighter (Battle Master)"


class KnowYourEnemy(Feature):
    """Starting at 7th level, if you spend at least 1 minute observing or
    interacting with another creature outside combat, you can learn
    certain information about its capabilities compared to your
    own. The DM tells you if the creature is your equal, superior, or
    inferior in regard to two of the following characteristics of your
    choice:

    - Strength score
    - Dexterity score
    - Constitution score
    - Armor Class
    - Current hit points
    - Total class levels (if any)
    - Fighter class levels (if any)

    """

    name = "Know Your Enemy"
    source = "Fighter (Battle Master)"


class Relentless(Feature):
    """Starting at 15th level, when you roll initiative and have no superiority
    dice remaining, you regain 1 superiority die.

    """

    name = "Relentless"
    source = "Fighter (Battle Master)"


# Maneuvers
class Maneuver(Feature):
    """
    A generic Maneuver
    """

    name = "Maneuver"
    source = "Fighter Maneuver (Battle Master)"


class CommandersStrike(Maneuver):
    """When you take the Attack action on your turn, you can forgo one of your
    attacks and use a bonus action to direct one of your companions to
    strike. When you do so, choose a friendly creature who can see or hear you
    and expend one superiority die. That creature can immediately use its
    reaction to make one weapon attack, adding the superiority die to the
    attack's damage roll.

    """

    name = "Commander's Strike"


class DisarmingAttack(Maneuver):
    """When you hit a creature with a weapon attack, you can expend one
    superiority die to attempt to disarm the target, forcing it to drop one
    item o f your choice that it's holding. You add the superiority die to the
    attack's damage roll, and the target must make a Strength saving throw. On
    a failed save, it drops the object you choose. The object lands at its
    feet.

    """

    name = "Disarming Attack"


class DistractingStrike(Maneuver):
    """When you hit a creature with a weapon attack, you can expend one
    superiority die to distract the creature, giving your allies an
    opening. You add the superiority die to the attack's damage roll. The next
    attack roll against the target by an attacker other than you has advantage
    if the attack is made before the start of your next turn.

    """

    name = "Distracting Strike"


class EvasiveFootwork(Maneuver):
    """When you move, you can expend one superiority die, rolling the die and
    adding the number rolled to your AC until you stop moving.

    """

    name = "Evasive Footwork"


class FeintingAttack(Maneuver):
    """You can expend one superiority die and use a bonus action on your turn to
    feint, choosing one creature within 5 feet of you as your target. You have
    advantage on your next attack roll against that creature. If that attack
    hits, add the superiority die to the attack's damage roll.

    """

    name = "Feinting Attack"


class GoadingAttack(Maneuver):
    """When you hit a creature with a weapon attack, you can expend one
    superiority die to attempt to goad the target into attacking you. You add
    the superiority die to the attack's damage roll, and the target must make a
    W isdom saving throw. On a failed save, the target has disadvantage on all
    attack rolls against targets other than you until the end of your next
    turn.

    """

    name = "Goading Attack"


class LungingAttack(Maneuver):
    """When you make a melee weapon attack on your turn, you can expend one
    superiority die to increase your reach for that attack by 5 feet. If you
    hit, you add the superiority die to the attack's damage roll.

    """

    name = "Lunging Attack"


class ManeuveringAttack(Maneuver):
    """When you hit a creature with a weapon attack, you can expend one
    superiority die to maneuver one o f your comrades into a more advantageous
    position. You add the superiority die to the attack's damage roll, and you
    choose a friendly creature who can see or hear you. That creature can use
    its reaction to move up to half its speed without provoking opportunity
    attacks from the target of your attack.

    """

    name = "Maneuvering Attack"


class MenacingAttack(Maneuver):
    """When you hit a creature with a weapon attack, you can expend one
    superiority die to attempt to frighten the target. You add the superiority
    die to the attack's damage roll, and the target must make a Wisdom saving
    throw. On a failed save, it is frightened of you until the end o f your
    next turn.

    """

    name = "Menacing Attack"


class Parry(Maneuver):
    """When another creature damages you with a melee attack, you can use your
    reaction and expend one superiority die to reduce the damage by the number
    you roll on your superiority die + your Dexterity modifier.

    """

    name = "Parry"


class PrecisionAttack(Maneuver):
    """When you make a weapon attack roll against a creature, you can expend one
    superiority die to add it to the roll. You can use this maneuver before or
    after making the attack roll, but before any effects of the attack are
    applied

    """

    name = "Precision Attack"


class PushingAttack(Maneuver):
    """When you hit a creature with a weapon attack, you can expend one
    superiority die to attempt to drive the target back. You add the
    superiority die to the attack's damage roll, and if the target is Large or
    smaller, it must make a Strength saving throw. On a failed save, you push
    the target up to 15 feet away from you.

    """

    name = "Pushing Attack"


class Rally(Maneuver):
    """On your turn, you can use a bonus action and expend one superiority die to
    bolster the resolve of one of your companions. When you do so, choose a
    friendly creature who can see or hear you. That creature gains temporary
    hit points equal to the superiority die roll + your Charisma modifier.

    """

    name = "Rally"


class Riposte(Maneuver):
    """When a creature misses you with a melee attack, you can use your reaction
    and expend one superiority die to make a melee weapon attack against the
    creature. If you hit, you add the superiority die to the attack's damage
    roll

    """

    name = "Riposte"


class SweepingAttack(Maneuver):
    """When you hit a creature with a melee weapon attack, you can expend one
    superiority die to attempt to damage another creature with the same
    attack. Choose another creature within 5 feet of the original target and
    within your reach. If the original attack roll would hit the second
    creature, it takes damage equal to the number you roll on your superiority
    die. The damage is of the same type dealt by the original attack. Trip

    """

    name = "Sweeping Attack"


class TripingAttack(Maneuver):
    """When you hit a creature with a weapon attack, you can expend one
    superiority die to attempt to knock the target down. You add the
    superiority die to the attack's damage roll, and if the target is Large or
    smaller, it must make a Strength saving throw. On a failed save, you knock
    the target prone

    """

    name = "Triping Attack"


# Eldritch Knight
class EldritchKnightSpellcasting(Feature):
    """You know three 1st-level wizard spells of your choice, two of which you
    must choose from the abjuration and evocation spells on the wizard spell
    list.

    The Spells Known column of the Eldritch Knight Spellcasting table shows
    when you learn more wizard spells of 1st level or higher. Each of these
    spells must be an abjuration or evocation spell of your choice, and must be
    of a level for which you have spell slots. For instance, when you reach 7th
    level in this class, you can learn one new spell of 1st or 2nd level.

    The spells you learn at 8th, 14th, and 20th level can come from any school
    of magic.

    Whenever you gain a level in this class, you can replace one of the wizard
    spells you know with another spell o f your choice from the wizard spell
    list. The new spell must be of a level for which you have spell slots, and
    it must be an abjuration or evocation spell, unless you're replacing the
    spell you gained at 8th, 14th, or 20th level.

    """

    name = "Spellcasting"
    source = "Fighter (Eldritch Knight)"


class WeaponBond(Feature):
    """At 3rd level, you learn a ritual that creates a magical bond between
    yourself and one weapon. You perform the ritual over the course of 1 hour,
    which can be done during a short rest. The weapon must be within your reach
    throughout the ritual, at the conclusion of which you touch the weapon and
    forge the bond.

    Once you have bonded a weapon to yourself, you can't be disarmed of that
    weapon unless you are incapacitated. If it is on the same plane of
    existence, you can summon that weapon as a bonus action on your turn,
    causing it to teleport instantly to your hand.

    You can have up to two bonded weapons, but can summon only one at a time
    with your bonus action. If you attempt to bond with a third weapon, you
    must break the bond with one of the other two.

    """

    name = "Weapon Bond"
    source = "Fighter (Eldritch Knight)"


class WarMagic(Feature):
    """Beginning at 7th level, when you use your action to cast a cantrip, you can
    make one weapon attack as a bonus action.

    """

    name = "War Magic"
    source = "Fighter (Eldritch Knight)"


class EldritchStrike(Feature):
    """At 10th level, you learn how to make your weapon strikes undercut a
    creature's resistance to your spells. When you hit a creature with a weapon
    attack, that creature has disadvantage on the next saving throw it makes
    against a spell you cast before the end of your next turn.

    """

    name = "Eldritch Strike"
    source = "Fighter (Eldritch Knight)"


class ArcaneCharge(Feature):
    """At 15th level, you gain the ability to teleport up to 30 feet to an
    unoccupied space you can see when you use your Action Surge. You can
    teleport before or after the additional action.

    """

    name = "Arcane Charge"
    source = "Fighter (Eldritch Knight)"


class ImprovedWarMagic(Feature):
    """Starting at 18th level, when you use your action to cast a spell, you can
    make one weapon attack as a bonus action.

    """

    name = "Improved War Magic"
    source = "Fighter (Eldritch Knight)"


# Purple Dragon Knight
class RallyingCry(Feature):
    """When you choose this archetype at 3rd level, you learn how to inspire your
    allies to fight on past their injuries. When you use your Second Wind
    feature, you can choose up to three creatures within 60 feet of you that
    are allied with you. Each one regains hit points equal to your fighter
    level, provided that the creature can see or hear you

    """

    name = "Rallying Cry"
    source = "Fighter (Purple Dragon Knight)"


class RoyalEnvoy(Feature):
    """A Purple Dragon knight serves as an envoy of the Cormyrean crown. Knights
    of high standing are expected to conduct themselves with grace.

    At 7th level, you gain proficiency in the Persuasion skill. If you are
    already proficient in it, you gain proficiency in one of the following
    skills of your choice: Animal Handling, Insight, Intimidation, or
    Performance.

    Your proficiency bonus is doubled for any ability check you make that uses
    Persuasion. You receive this benefit regardless of the skill proficiency
    you gain from this feature

    """

    name = "Royal Envoy"
    source = "Fighter (Purple Dragon Knight)"

    def __init__(self, owner=None):
        super().__init__(owner=owner)
        if "persuasion" not in self.owner.skill_proficiencies:
            self.owner.skill_proficiencies.append("persuasion")
        if "persuasion" not in self.owner.skill_expertise:
            self.owner.skill_expertise.append("persuasion")


class InspiringSurge(Feature):
    """Starting at 10th level, when you use your Action Surge feature , you can
    choose one creature within 60 feet of you that is allied with you. That
    creature can make one melee or ranged weapon attack with its reaction,
    provided that it can see or hear you. Starting at 18th level, you can choose
    two allies within 60 feet of you, rather than one.

    """

    name = "Inspiring Surge"
    source = "Fighter (Purple Dragon Knight)"


class Bulwark(Feature):
    """Beginning at 15th level, you can extend the benefit of your Indomitable
    feature to an ally. When you decide to use Indomitable to reroll an
    Intelligence, a Wisdom, or a Charisma saving throw and you aren't
    incapacitated, you can choose one ally within 60 feet of you that also
    failed its saving throw against the same effect. If that creature can see
    or hear you, it can reroll its saving throw and must use the new roll

    """

    name = "Bulwark"
    source = "Fighter (Purple Dragon Knight)"


# Arcane Archer
class ArcaneArcherLore(Feature):
    """At 3rd level, you learn magical theory or some of the secrets of
    nature-typical for practitioners of this elven martial tradition. You
    choose to gain proficiency in ei- ther the Arcana or the Nature skill, and
    you choose to learn either the prestidigr'tation or the drufdcraft cantrip

    """

    name = "Arcane Archer Lore"
    source = "Fighter (Arcane Archer)"


class ArcaneShot(Feature):
    """At 3rd level, you learn to unleash special magical effects with some of
    your shots. When you gain this feature, you learn two Arcane Shot options
    of your choice (see "Arcane Shot Options" below).

    Add all Arcane Shot options under "features" in your .py file

    Once per turn when you fire a magic arrow from a shortbow or longbow as
    part of the Attack action, you can apply one of your Arcane Shot options to
    that arrow. You decide to use the option when the arrow hits a creature,
    unless the option doesn't involve an attack roll. You have two uses of this
    ability, and you regain all expended uses of it when you finish a short or
    long rest.

    You gain an additional Arcane Shot option of your choice when you reach
    certain levels in this class: 7th, 10th, 15th, and 18th level. Each option
    also improves when you become an 18th-level fighter

    If an option requires a saving throw, your Arcane Shot save DC equals 8 +
    your proficiency bonus + your Intelligence modifier

    """

    name = "Arcane Shot (2x/SR)"
    source = "Fighter (Arcane Archer)"


class MagicArrow(Feature):
    """At 7th level, you gain the ability to infuse arrows with magic. Whenever
    you fire a nonmagical arrow from a shortbow or longbow, you can make it
    magical for the purpose of overcoming resistance and immunity to non-
    magical attacks and damage. The magic fades from the arrow immediately
    after it hits or misses its target.

    """

    name = "Magic Arrow"
    source = "Fighter (Arcane Archer)"


class CurvingShot(Feature):
    """At 7th level, you learn how to direct an errant arrow toward a new
    target. When you make an attack roll with a magic arrow and miss, you can
    use a bonus action to reroll the attack roll against a different target
    within 60 feet of the original target

    """

    name = "Curving Shot"
    source = "Fighter (Arcane Archer)"


class EverReadyShot(Feature):
    """Starting at 15th level, your magical archery is available whenever battle
    starts. If you roll initiative and have no uses of Arcane Shot remaining,
    you regain one use of it.

    """

    name = "Ever-Ready Shot"
    source = "Fighter (Arcane Archer)"


class BanishingArrow(Feature):
    """You use abjuration magic to try to temporarily banish your target to a
    harmless location in the Feywild. The creature hit by the arrow must also
    succeed on a Charisma saving throw or be banished. While banished in this
    way, the target's speed is 0, and it is incapacitated. At the end of its
    next turn, the target reappears in the Space it vacated or in the nearest
    unoccupied space if that space is occupied.

    After you reach 18th level in this class, a target also takes 2d6 force
    damage when the arrow hits it.

    """

    name = "Arcane Shot: Banishing Arrow"
    source = "Fighter (Arcane Archer)"


class BeguilingArrow(Feature):
    """Your enchantment magic causes this arrow to temporarily beguile its
    target. The creature hit by the arrow takes an extra 2d6 psychic damage,
    and choose one ofyour allies within 30 feet of the target. The target must
    succeed on a Wisdom saving throw, or it is charmed by the chosen ally until
    the start ofyour next turn. This effect ends early if the chosen ally
    attacks the charmed target, deals damage to it, or forces it to make
    a saving throw. The psychic damage increases to 4d6 when you reach 18th
    level in this class.

    """

    name = "Arcane Shot: Beguiling Arrow"
    source = "Fighter (Arcane Archer)"


class BurstingArrow(Feature):
    """You imbue your arrow with force energy drawn from the school of
    evocation. The energy detonates after your attack. Immediately after the
    ar- row hits the creature, the target and all other creatures within 10
    feet of it take 2d6 force damage each. The force damage increases to 4d6
    when you reach 18th level in this class

    """

    name = "Arcane Shot: Bursting Arrow"
    source = "Fighter (Arcane Archer)"


class EnfeeblingArrow(Feature):
    """You weave necromantic magic into your arrow. The creature hit by the arrow
    takes an extra 2d6 necrotic damage. The target must also succeed on a
    Constitution saving throw, or the damage dealt by its weapon attacks is
    halved until the start of your next turn. The necrotic damage increases to
    4d6 when you reach 18th level in this class.

    """

    name = "Arcane Shot: Enfeebling Arrow"
    source = "Fighter (Arcane Archer)"


class GraspingArrow(Feature):
    """When this arrow strikes its target, conjuration magic creates grasping,
    poisonous brams bles, which wrap around the target. The creature hit by the
    arrow takes an extra 2(16 poison damage, its speed is reduced by 10 feet,
    and it takes 2d6 slashing dam- age the first time on each turn it moves 1
    foot or more without teleporting. The target or any creature that can reach
    it can use its action to remove the brambles with a successful Strength
    (Athletics) check against your Arcane Shot save DC. Otherwise, the
    brambles last for 1 minute or until you use this option again. The poison
    damage and slashing damage both increase to 4d6 when you reach 18th level
    in this class

    """

    name = "Arcane Shot: Grasping Arrow"
    source = "Fighter (Arcane Archer)"


class PiercingArrow(Feature):
    """You use transmutation magic to give your arrow an ethereal quality. When
    you use this option, you don't make an attack roll for the attack. Instead,
    the arrow shoots forward in a line, which is 1 foot wide and 30 feet long,
    before disappearing. The arrow passes harmlessly through objects, ignoring
    cover. Each creature in that line must make a Dexterity saving throw. On a
    failed save, a creature takes damage as if it were hit by the arrow, plus
    an extra 1d6 piercing damage. On a successful save, a target takes half as
    much damage. The piercing damage increases to 2d6 when you reach 18th level
    in this class.

    """

    name = "Arcane Shot: Piercing Arrow"
    source = "Fighter (Arcane Archer)"


class SeekingArrow(Feature):
    """Using divination magic, you grant your arrow the ability to seek out a
    target. When you use this option, you don't make an attack roll for the
    attack. Instead, choose one creature you have seen in the past minute. The
    arrow flies toward that creature, moving around corners if necessary and
    ignoring three-quarters cover and half cover. If the target is within the
    weapon's range and there is a path large enough for the arrow to travel to
    the target, the target must make a Dexterity saving throw. Otherwise, the
    arrow disappears after traveling as far as it can. On a failed save, the
    target takes damage as if it were hit by the arrow, plus an extra 1d6 force
    damage, and you learn the target's current location. On a successful save,
    the target takes half as much damage, and you don't learn its location. The
    force damage increases to 2d6 when you reach 18th level in this class.

    """

    name = "Arcane Shot: Seeking Arrow"
    source = "Fighter (Arcane Archer)"


class ShadowArrow(Feature):
    """You weave illusion magic into your arrow, causing it to occlude your fees
    vision with shadows. The creature hit by the arrow takes an extra 2d6
    psychic damage, and it must succeed on a Wisdom saving throw or be unable
    to see anything farther than 5 feet away until the start ofyour next
    turn. The psychic damage increases to 4d6 when you reach 18th level in this
    class

    """

    name = "Shadow Arrow"
    source = "Fighter (Arcane Archer)"


# Cavalier
class BonusProficiencyCavalier(Feature):
    """When you choose this archetype at 3rd level, you gain proficiency in one of
    the following skills of your choice: Animal Handling, History, Insight,
    Performance, or Persuasion. Alternatively, you learn one language of your
    choice.

    """

    name = "Bonus Proficiency"
    source = "Fighter (Cavalier)"


class BornToTheSaddle(Feature):
    """Starting at 3rd level, your mastery as a rider becomes apparent. You have
    advantage on saving throws made to avoid falling off your mount. Ifyou fall
    off your mount and descend no more than 10 feet, you can land on your feet
    if you're not incapacitated. Finally, mounting or dismounting a creature
    costs you only 5 feet of movement, rather than half your speed.

    """

    name = "Born to the Saddle"
    source = "Fighter (Cavalier)"


class UnwaveringMark(Feature):
    """Starting at 3rd level, you can menace your foe-s, foiling their attacks and
    punishing them for harming others. When you hit a creature with a melee
    weapon attack, you can mark the creature until the end of your next
    turn. This effect ends early ifyou are incapacitated or you die, or if
    someone else marks the creature.

    While it is within 5 feet of you, a creature marked by you has disadvantage
    on any attack roll that doesn't target you. In addition, if a creature
    marked by you deals damage to anyone other than you, you can make a special
    melee weapon attack against the marked creature as a bonus action on your
    next turn. You have advantage on the attack roll, and if it hits, the
    attacks weapon deals extra damage to the target equal to half your fighter
    level.

    Regardless of the number of creatures you mark, you can make this special
    attack a number of times equal to your Strength modifier (minimum of once),
    and you regain all expended uses of it when you finish a long rest

    """

    _name = "Unwavering Mark"
    source = "Fighter (Cavalier)"

    @property
    def name(self):
        num = max(1, self.owner.strength.modifier)
        return self._name + " ({:d}x/LR)".format(num)


class WardingManeuver(Feature):
    """At 7th level, you learn to fend off strikes directed at you, your mount, or
    other creatures nearby. If you or a creature you can see within 5 feet
    of you is hit by an attack, you can roll ld8 as a reaction if you're
    wielding a melee weapon or a shield. Roll the die, and add the number
    rolled to the target's AC against that attack. If the attack still hits,
    the target has resistance against the attack's damage.

    You can use this feature a number of times equal to your Constitution
    modifier (minimum of once), and you regain all expended uses of it when you
    finish a long rest

    """

    _name = "Warding Maneuver"
    source = "Fighter (Cavalier)"

    @property
    def name(self):
        num = max(1, self.owner.constitution.modifier)
        return self._name + " ({:d}x/LR)".format(num)


class HoldTheLine(Feature):
    """At 10th level, you become a master of locking down your enemies. Creatures
    provoke an opportunity attack from you when they move 5 feet or more while
    within your reach, and if you hit a creature with an opportunity attack,
    the target's speed is reduced to 0 until the end of the current turn.

    """

    name = "Hold the Line"
    source = "Fighter (Cavalier)"


class FerociousCharger(Feature):
    """Starting at 15th level, you can run down your foes, whether you're mounted
    or not. If you move at least 10 feet in a straight line right before
    attacking a creature and you hit it with the attack, that target must
    succeed on a Strength saving throw (DC 8 + your proficiency bonus + your
    Strength modifier) or be knocked prone. You can use this feature only once
    on each of your turns

    """

    name = "Ferocious Charger"
    source = "Fighter (Cavalier)"


class VigilantDefender(Feature):
    """Starting at 18th level, you respond to danger with extraordinary
    vigilance. In combat, you get a special reaction that you can take once on
    every creature's turn, except your turn. You can use this special reaction
    only to make an opportunity attack, and you can't use it on the same turn
    that you take your normal reaction

    """

    name = "Vigiland Defender"
    source = "Fighter (Cavalier)"


# Samurai
class BonusProficiencySamurai(Feature):
    """When you choose this archetype at 3rd level, you gain proficiency in one of
    the following skills ofyour choice: History, Insight, Performance, or
    Persuasion. Alterna- tively, you learn one language of your choice.

    """

    name = "Bonus Proficiency"
    source = "Fighter (Samurai)"


class FightingSpirit(Feature):
    """Starting at 3rd level, your intensity in battle can shield you and help you
    strike true. As a bonus action on your turn, you can give yourself
    advantage on weapon attack rolls until the end of the current turn. When
    you do so, you also gain 5 temporary hit points. The number of temporary
    hit points increases when you reach certain levels in this class,
    increasing to 10 at 10th level and 15 at 15th level. You can use this
    feature three times, and you regain all expended uses of it when you finish
    a long rest.

    """

    name = "Fighting Spirit (3x/LR)"
    source = "Fighter (Samurai)"


class ElegantCourtier(Feature):
    """Starting at 7th level, your discipline and attention to de- tail allow you
    to excel in social situations. Whenever you make a Charisma (Persuasion)
    check, you gain a bonus to the check equal to your Wisdom modifier. Your
    self-control also causes you to gain proficiency in Wisdom saving
    throws. If you already have this proficiency, you instead gain proficiency
    in Intelligence or Charisma saving throws (your choice).

    """

    name = "Elegant Courtier"
    source = "Fighter (Samurai)"
    needs_implementation = True


class TirelessSpirit(Feature):
    """Starting at 10th level, when you roll initiative and have no uses of
    Fighting Spirit remaining, you regain one use.

    """

    name = "Tireless Spirit"
    source = "Fighter (Samurai)"


class RapidStrike(Feature):
    """Starting at 15th level, you learn to trade accuracy for swift strikes. If
    you take the Attack action on your turn and have advantage on an attack
    roll against one of the targets, you can forgo the advantage for that roll
    to make an additional weapon attack against that target, as part of the
    same action. You can do so no more than once per turn

    """

    name = "Rapid Strike"
    source = "Fighter (Samurai)"


class StrengthBeforeDeath(Feature):
    """Starting at 18th level, your fighting spirit can delay the grasp of
    death. If you take damage that reduces you to 0 hit points and doesn't kill
    you outright, you can use your reaction to delay falling unconscious, and
    you can immediately take an extra turn, interrupting the current
    turn. While you have 0 hit points during that extra turn, taking damage
    causes death saving throw failures as normal, and three death saving throw
    failures can still kill you. When the extra turn ends, you fall unconscious
    ifyou still have 0 hit points. Once you use this feature, you can't use it
    again until you finish a long rest.

    """

    name = "Strength Before Death"
    source = "Fighter (Samurai)"


# Gunslinger
class Gunsmith(Feature):
    """Upon choosing this archetype at 3rd level, you gain proficiency with
    Tinker's Tools. You may use them to craft ammunition at half the cost,
    repair damaged firearms, or even draft and create new ones (DM's
    discretion). Some extremely experimental and intricate firearms are only
    available through crafting.

    """

    name = "Gunsmith"
    source = "Fighter (Gunslinger)"


class AdeptMarksman(Feature):
    """When you choose this archetype at 3rd level, you learn to perform powerful
    trick shots to disable or damage your opponents using your firearms.

    **Trick Shots**: You learn two trick shots of your choice, which are
    detailed under "Trick Shots" below.  If you have not already, add them by
    name to "features" in your character's .py file.

    Many maneuvers enhance an attack in some way. Each use of a trick shot must
    be declared before the attack roll is made. You can use only one trick shot
    per attack.

    You learn an additional trick shot of your choice at 7th, 10th, 15th, and
    18th level. Each time you learn a new trick shot, you can also replace one
    trick shot you know with a different one.

    Grit. You gain a number of grit points equal to your Wisdom modifier
    (minimum of 1). You regain 1 expended grit point each time you roll a 20 on
    the d20 roll for an attack with a firearm, or deal a killing blow with a
    firearm to a creature of significant threat (DM's discretion). You regain
    all expended grit points after a short or long rest.

    Saving Throws. Some of your trick shots require your targets to make a
    saving throw to resist the trick shot's effects. The saving throw DC is
    calculated as follows:

    Trick Shot save DC = 8 + your proficiency bonus + your Dexterity modifier

    Firearm Properties Firearms are a new and volatile technology, and as such
    bring their own unique set of weapon properties. Some properties are
    followed by a number, and this number signifies an element of that property
    (outlined below). These properties replace the optional ones presented in
    the Dungeon Master's Guide. Firearms are ranged weapons.

    Reload. The weapon can be fired a number of times equal to its Reload score
    before you must spend 1 attack or 1 action to reload. You must have one
    free hand to reload a firearm.

    Misfire. Whenever you make an attack roll with a firearm, and the dice roll
    is equal to or lower than the weapon's Misfire score, the weapon
    misfires. The attack misses, and the weapon cannot be used again until you
    spend an action to try and repair it. To repair your firearm, you must make
    a successful Tinker's Tools check (DC equal to 8 + misfire score). If your
    check fails, the weapon is broken and must be mended out of combat at a
    quarter of the cost of the firearm. Creatures who use a firearm without
    being proficient increase the weapon's misfire score by 1.

    Explosive. Upon a hit, everything within 5 ft of the target must make a
    Dexterity saving throw (DC equal to 8 + your proficiency bonus + your
    Dexterity modifier) or suffer 1d8 fire damage. If the weapon misses, the
    ammunition fails to detonate, or bounces away harmlessly before doing so.

    Ammunition All firearms require ammunition to make an attack, and due to
    their rare nature, ammunition may be near impossible to find or
    purchase. However, if materials are gathered, you can craft ammunition
    yourself using your Tinker's Tools at half the cost. Each firearm uses its
    own unique ammunition and is generally sold or crafted in batches listed
    below next to the price.

    """

    name = "Adept Marksman"
    source = "Fighter (Gunslinger"


class QuickDraw(Feature):
    """When you reach 7th level, you add your proficiency bonus to your
    initiative. You can also stow a firearm, then draw another firearm as a
    single object interaction on your turn.

    """

    name = "Quick Draw"
    source = "Fighter (Gunslinger)"


class RapidRepair(Feature):
    """Upon reaching 10th level, you learn how to quickly attempt to fix a jammed
    gun. You can spend a grit point to attempt to repair a misfired (but not
    broken) firearm as a bonus action.

    """

    name = "Rapid Repair"
    source = "Fighter (Gunslinger)"


class LightningReload(Feature):
    """Starting at 15th level, you can reload any firearm as a bonus action."""

    name = "Lightning Repaid"
    source = "Fighter (Gunslinger)"


class ViciousIntent(Feature):
    """At 18th level, your firearm attacks score a critical hit on a roll of 19-20,
    and you regain a grit point on a roll of 19 or 20 on a d20 attack roll.

    """

    name = "Vicious Intent"
    source = "Fighter (Gunslinger)"


class HemorrhagingCritical(Feature):
    """Upon reaching 18th level, whenever you score a critical hit on an attack
    with a firearm, the target additionally suffers half of the damage from the
    attack at the end of its next turn.

    """

    name = "Hemorrhaging Critical"
    source = "Fighter (Gunslinger)"


class BullyingShot(Feature):
    """You can use the powerful blast and thundering sound of your firearm to shake
    the resolve of a creature. You can expend one grit point while making a
    Charisma (Intimidation) check to gain advantage on the roll.

    """

    name = "Bullying Shot"
    source = "Gunslinger (Trick Shot)"


class DazingShot(Feature):
    """When you make a firearm attack against a creature, you can expend one grit
    point to attempt to dizzy your opponent. On a hit, the creature suffers
    normal damage and must make a Constitution saving throw or suffer
    disadvantage on attacks until the end of their next turn.

    """

    name = "Dazing Shot"
    source = "Gunslinger (Trick Shot)"


class DeadeyeShot(Feature):
    """When you make a firearm attack against a creature, you can expend one grit
    point to gain advantage on the attack roll.

    """

    name = "Deadeye Shot"
    source = "Gunslinger (Trick Shot)"


class DisarmingShot(Feature):
    """When you make a firearm attack against a creature, you can expend one grit
    point to attempt to shoot an object from their hands. On a hit, the
    creature suffers normal damage and must succeed on a Strength saving throw
    or drop 1 held object of your choice and have that object be pushed 10 feet
    away from you.

    """

    name = "Disarming Shot"
    source = "Gunslinger (Trick Shot)"


class ForcefulShot(Feature):
    """When you make a firearm attack against a creature, you can expend one grit
    point to attempt to trip them up and force them back. On a hit, the
    creature suffers normal damage and must succeed on a Strength saving throw
    or be pushed 15 feet away from you.

    """

    name = "Forceful Shot"
    source = "Gunslinger (Trick Shot)"


class PiercingShot(Feature):
    """When you make a firearm attack against a creature, you can expend one grit
    point to attempt to fire through multiple opponents. The initial attack
    gains a +1 to the firearm's misfire score. On a hit, the creature suffers
    normal damage and you make an attack roll with disadvantage against every
    creature in a line directly behind the target within your first range
    increment. Only the initial attack can misfire.

    """

    name = "Piercing Shot"
    source = "Gunslinger (Trick Shot)"


class WingingShot(Feature):
    """When you make a firearm attack against a creature, you can expend one grit
    point to attempt to topple a moving target. On a hit, the creature suffers
    normal damage and must make a Strength saving throw or be knocked prone.

    """

    name = "Winging Shot"
    source = "Gunslinger (Trick Shot)"


class ViolentShot(Feature):
    """When you make a firearm attack against a creature, you can expend one or
    more grit points to enhance the volatility of the attack. For each grit
    point expended, the attack gains a +2 to the firearm's misfire score. If
    the attack hits, you can roll one additional weapon damage die per grit
    point spent when determining the damage.

    """

    name = "Violent Shot"
    source = "Gunslinger (Trick Shot)"



from collections import defaultdict

from dungeonsheets import features as feats
from dungeonsheets import spells, weapons
from dungeonsheets.content_registry import default_content_registry


default_content_registry.add_module(__name__)


class Race:
    name = "Unknown"
    size = "medium"
    speed = 30
    owner = None
    languages = ("Common",)
    proficiencies_text = tuple()
    weapon_proficiences = tuple()
    skill_proficiencies = ()
    skill_choices = ()
    num_skill_choices = 0
    features = tuple()
    features_by_level = defaultdict(list)
    strength_bonus = 0
    dexterity_bonus = 0
    constitution_bonus = 0
    intelligence_bonus = 0
    wisdom_bonus = 0
    charisma_bonus = 0
    hit_point_bonus = 0
    spells_known = ()

    def __init__(self, owner=None):
        self.owner = owner
        cls = type(self)
        # Instantiate the features
        self.features = tuple([f(owner=self.owner) for f in cls.features])
        self.features_by_level = defaultdict(list)
        for i in range(1, 21):
            self.features_by_level[i] = [
                f(owner=self.owner) for f in cls.features_by_level[i]
            ]
        self.spells_known = [S() for S in cls.spells_known]

    @property
    def spells_prepared(self):
        return self.spells_known

    def __str__(self):
        return self.name

    def __repr__(self):
        return '"{:s}"'.format(self.name)


# Dwarves
class _Dwarf(Race):
    name = "Dwarf"
    size = "medium"
    speed = 25
    languages = ("Common", "Dwarvish")
    constitution_bonus = 2
    proficiencies_text = ("battleaxes", "handaxes", "throwing hammers", "warhammers")
    weapon_proficiences = (
        weapons.Battleaxe,
        weapons.Handaxe,
        weapons.ThrowingHammer,
        weapons.Warhammer,
    )
    features = (feats.Darkvision, feats.DwarvenResilience, feats.Stonecunning)


class HillDwarf(_Dwarf):
    name = "Hill Dwarf"
    wisdom_bonus = 1
    hit_point_bonus = 1
    features = _Dwarf.features + (feats.DwarvenToughness,)


class MountainDwarf(_Dwarf):
    name = "Mountain Dwarf"
    strength_bonus = 2


# Elves
class _Elf(Race):
    name = "Elf"
    size = "medium"
    speed = 30
    dexterity_bonus = 2
    skill_proficiencies = ("perception",)
    languages = ("Common", "Elvish")
    features = (feats.Darkvision, feats.FeyAncestry, feats.Trance)


class HighElf(_Elf):
    name = "High Elf"
    weapon_proficiencies = (
        weapons.Longsword,
        weapons.Shortsword,
        weapons.Shortbow,
        weapons.Longbow,
    )
    proficiencies_text = ("longswords", "shortswords", "shortbows", "longbows")
    intelligence_bonus = 1
    languages = ("Common", "Elvish", "[choose one]")
    features = _Elf.features + (feats.ElfCantrip,)


class WoodElf(_Elf):
    name = "Wood Elf"
    weapon_proficiencies = (
        weapons.Longsword,
        weapons.Shortsword,
        weapons.Shortbow,
        weapons.Longbow,
    )
    proficiencies_text = ("longswords", "shortswords", "shortbows", "longbows")
    wisdom_bonus = 1
    speed = 35
    features = _Elf.features + (feats.MaskOfTheWild,)


class DarkElf(_Elf):
    name = "Dark Elf"
    weapon_proficiencies = (weapons.Rapier, weapons.Shortsword, weapons.HandCrossbow)
    proficiencies_text = ("rapiers", "shortswords", "hand crossbows")
    charisma_bonus = 1
    features = (
        feats.SuperiorDarkvision,
        feats.FeyAncestry,
        feats.Trance,
        feats.SunlightSensitivity,
        feats.DrowMagic,
    )
    spells_known = (spells.DancingLights,)


Drow = DarkElf


# Halflings
class _Halfling(Race):
    name = "Halfling"
    size = "small"
    speed = 25
    dexterity_bonus = 2
    languages = ("Common", "Halfling")
    features = (feats.Lucky, feats.Brave, feats.HalflingNimbleness)


class LightfootHalfling(_Halfling):
    name = "Lightfoot Halfling"
    charisma_bonus = 1
    features = _Halfling.features + (feats.NaturallyStealthy,)


class StoutHalfling(_Halfling):
    name = "Stout Halfling"
    constitution_bonus = 1
    features = _Halfling.features + (feats.StoutResilience,)


# Humans
class Human(Race):
    name = "Human"
    size = "medium"
    speed = 30
    strength_bonus = 1
    dexterity_bonus = 1
    constitution_bonus = 1
    intelligence_bonus = 1
    wisdom_bonus = 1
    charisma_bonus = 1
    languages = ("Common", "[choose one]")


class Rashemi(Human):
    name = "Rashemi"


# Dragonborn
class Dragonborn(Race):
    name = "Dragonborn"
    size = "medium"
    speed = 30
    strength_bonus = 2
    charisma_bonus = 1
    languages = ("Common", "Draconic")
    features = (feats.DraconicAncestry, feats.BreathWeapon, feats.DraconicResistance)


# Gnomes
class _Gnome(Race):
    name = "Gnome"
    size = "small"
    speed = 25
    intelligence_bonus = 2
    languages = ("Common", "Gnomish")
    features = (feats.Darkvision, feats.GnomeCunning)


class ForestGnome(_Gnome):
    name = "Forest Gnome"
    dexterity_bonus = 1
    features = _Gnome.features + (feats.NaturalIllusionist, feats.SpeakWithSmallBeasts)
    spells_known = (spells.MinorIllusion,)


class RockGnome(_Gnome):
    name = "Rock Gnome"
    constitution_bonus = 1
    features = _Gnome.features + (feats.ArtificersLore, feats.Tinker)


class DeepGnome(_Gnome):
    name = "Deep Gnome"
    dexterity_bonus = 1
    languages = ("Common", "Gnomish", "Undercommon")
    features = (feats.SuperiorDarkvision, feats.GnomeCunning, feats.StoneCamouflage)


# Half-elves
class HalfElf(Race):
    name = "Half-Elf"
    size = "medium"
    speed = 30
    charisma_bonus = 2
    skill_choices = (
        "acrobatics",
        "animal handling",
        "arcana",
        "athletics",
        "deception",
        "history",
        "insight",
        "intimidation",
        "investigation",
        "medicine",
        "nature",
        "perception",
        "performance",
        "persuasion",
        "religion",
        "sleight of hand",
        "stealth",
        "survival",
    )
    num_skill_choices = 2
    languages = ("Common", "Elvish", "[choose one]")
    features = (feats.Darkvision, feats.FeyAncestry)


# Half-Orcs
class HalfOrc(Race):
    name = "Half-Orc"
    size = "medium"
    speed = 30
    strength_bonus = 2
    constitution_bonus = 1
    skill_proficiencies = ("intimidation",)
    languages = ("Common", "Orc")
    features = (feats.Darkvision, feats.RelentlessEndurance, feats.SavageAttacks)


# Tieflings
class Tiefling(Race):
    name = "Tiefling"
    size = "medium"
    speed = 30
    intelligence_bonus = 1
    charisma_bonus = 2
    languages = ("Common", "Infernal")
    features = (feats.Darkvision, feats.HellishResistance, feats.InfernalLegacy)


# Aasimar
class _Aasimar(Race):
    name = "Aasimar"
    size = "medium"
    speed = 30
    charisma_bonus = 2
    languages = ("Common", "Celestial")
    features = (
        feats.Darkvision,
        feats.CelestialResistance,
        feats.HealingHands,
        feats.LightBearer,
    )
    spells_known = (spells.Light,)


# Protector Aasimar
class ProtectorAasimar(_Aasimar):
    name = "Protector Aasimar"
    wisdom_bonus = 1
    features_by_level = defaultdict(list)
    features_by_level[3] += [feats.RadiantSoul]


# Fallen Aasimar
class ScourgeAasimar(_Aasimar):
    name = "Scourge Aasimar"
    constitution_bonus = 1
    features_by_level = defaultdict(list)
    features_by_level[3] += [feats.RadiantConsumption]


# Fallen Aasimar
class FallenAasimar(_Aasimar):
    name = "Fallen Aasimar"
    strength_bonus = 1
    features_by_level = defaultdict(list)
    features_by_level[3] += [feats.NecroticShroud]


# Firbolg
class Firbolg(Race):
    name = "Firbolg"
    size = "medium"
    speed = 30
    wisdom_bonus = 2
    strength_bonus = 1
    features = (
        feats.FirbolgMagic,
        feats.HiddenStep,
        feats.PowerfulBuild,
        feats.SpeechOfBeastAndLeaf,
    )
    languages = ("Common", "Elvish", "Giant")


# Goliath
class Goliath(Race):
    name = "Goliath"
    size = "Medium"
    speed = 30
    skill_proficiencies = ("athletics",)
    languages = ("Common", "Giant")
    features = (feats.StonesEndurance, feats.PowerfulBuild, feats.MountainBorn)


# Lizardfolk
class Lizardfolk(Race):
    name = "Lizardfolk"
    size = "medium"
    speed = """30 (30 swim)"""
    constitution_bonus = 2
    wisdom_bonus = 1
    languages = ("Common", "Draconic")
    weapon_proficiencies = (weapons.Bite,)
    proficiencies_text = ("bite",)
    features = (
        feats.CunningArtisan,
        feats.HoldBreath,
        feats.NaturalArmor,
        feats.HungryJaws,
    )
    skill_choices = ("animal handling", "nature", "perception", "stealth", "survival")

    def __init__(self, owner=None):
        super().__init__(owner=owner)
        self.owner.wield_weapon("bite")


# Kenku
class Kenku(Race):
    name = "Kenku"
    size = "medium"
    speed = 30
    dexterity_bonus = 2
    wisdom_bonus = 1
    languages = ("Common", "Auran")
    skill_choices = ("acrobatics", "deception", "stealth", "sleight of hand")
    num_skill_choices = 2
    features = (
        feats.ExpertForgery,
        feats.Mimicry,
    )


# Tabaxi
class Tabaxi(Race):
    name = "Tabaxi"
    size = "medium"
    dexterity_bonus = 2
    charisma_bonus = 1
    speed = "30 (20 climb)"
    languages = ("Common", "[Choose One]")
    weapon_proficiencies = (weapons.Claws,)
    proficiences_text = ("Claws",)
    skill_proficiencies = ("perception", "stealth")
    features = (
        feats.Darkvision,
        feats.FelineAgility,
    )

    def __init__(self, owner=None):
        super().__init__(owner=owner)
        self.owner.wield_weapon("claws")


# Triton
class Triton(Race):
    name = "Triton"
    size = "medium"
    strength_bonus = 1
    constitution_bonus = 1
    charisma_bonus = 1
    speed = "30 (30 swim)"
    features = (
        feats.Amphibious,
        feats.ControlAirAndWater,
        feats.EmissaryOfTheSea,
        feats.GuardiansOfTheDepths,
    )
    languages = ("Common", "Primordial")
    spells_known = (spells.FogCloud,)


# Aarakocra
class Aarakocra(Race):
    name = "Aarakocra"
    size = "medium"
    speed = "25 (50 fly)"
    dexterity_bonus = 2
    wisdom_bonus = 1
    languages = ("Common", "Aarakocra", "Auran")
    weapon_proficiencies = (weapons.Talons,)
    proficiences_text = ("Talons",)

    def __init__(self, owner=None):
        super().__init__(owner=owner)
        self.owner.wield_weapon("talons")


# Genasi
class _Genasi(Race):
    name = "Genasi"
    constitution_bonus = 2
    size = "medium"
    speed = 30
    languages = ("Common", "Primoridal")


class AirGenasi(_Genasi):
    name = "Air Genasi"
    dexterity_bonus = 1
    features = (feats.UnendingBreath, feats.MingleWithTheWind)


class EarthGenasi(_Genasi):
    name = "Earth Genasi"
    strength_bonus = 1
    features = (feats.EarthWalk, feats.MergeWithStone)


class FireGenasi(_Genasi):
    name = "Fire Genasi"
    intelligence_bonus = 1
    features = (feats.Darkvision, feats.FireResistance, feats.ReachToTheBlaze)


class WaterGenasi(_Genasi):
    name = "Water Genasi"
    wisdom_bonus = 1
    speed = "30 (30 swim)"
    features = (feats.AcidResistance, feats.Amphibious, feats.CallToTheWave)


# Eberron Races
class Kalashtar(Race):
    name = "Kalashtar"
    wisdom_bonus = 2
    charisma_bonus = 1
    size = "medium"
    speed = 30
    languages = (
        "Common",
        "Quori",
    )  # Not sure how to have a "+1 language of your choice" - naviabbot
    features = (
        feats.DualMind,
        feats.MentalDiscipline,
        feats.MindLink,
        feats.SeveredFromDreams,
    )


class BugBear(Race):
    name = "BugBear"
    strength_bonus = 2
    dexterity_bonus = 1
    size = "medium"
    speed = 30
    features = (
        feats.Darkvision,
        feats.LongLimbed,
        feats.PowerfulBuild,
        feats.SupriseAttack,
    )
    skill_proficiencies = ("stealth",)
    languages = ("Common", "Goblin")


class Goblin(Race):
    name = "Goblin"
    dexterity_bonus = 2
    constitution_bonus = 1
    size = "small"
    speed = 30
    features = (feats.Darkvision, feats.FuryOfTheSmall, feats.NimbleEscape)
    languages = ("Common", "Goblin")


class HobGoblin(Race):
    name = "HobGoblin"
    constitution_bonus = 2
    intelligence_bonus = 1
    size = "medium"
    speed = 30
    features = (feats.Darkvision, feats.SavingFace)
    proficiencies_text = ("light armor", "[Chose two martial melee weapons]")
    languages = ("Common", "Goblin")


class Kobold(Race):
    name = "Kobold"
    dexterity_bonus = 2
    strength_bonus = -2
    size = "small"
    speed = 30
    features = (
        feats.Darkvision,
        feats.PackTactics,
        feats.GrovelCowerAndBeg,
        feats.SunlightSensitivity,
    )
    languages = ("Common", "Draconic")


class Orc(Race):
    name = "Orc"
    strength_bonus = 2
    constitution_bonus = 1
    intelligence_bonus = -2
    size = "medium"
    speed = 30
    skill_proficiencies = ("intimidation",)
    features = (feats.Darkvision, feats.Aggressive, feats.PowerfulBuild)
    languages = ("Common", "Orc")


class PureBlood(Race):
    name = "Yuan-Ti Pureblood"
    charisma_bonus = 2
    intelligence_bonus = 1
    size = "medium"
    speed = 30
    features = (
        feats.Darkvision,
        feats.InnateSpellcasting,
        feats.MagicResistance,
        feats.PoisonImmunity,
    )
    spells_known = (spells.PoisonSpray,)
    languages = ("Common", "Abyssal", "Draconic")


PHB_races = [
    HillDwarf,
    MountainDwarf,
    HighElf,
    WoodElf,
    DarkElf,
    LightfootHalfling,
    StoutHalfling,
    Rashemi,
    Dragonborn,
    ForestGnome,
    RockGnome,
    HalfElf,
    HalfOrc,
    Tiefling,
    Human,
]

VOLO_races = [
    ProtectorAasimar,
    ScourgeAasimar,
    FallenAasimar,
    Firbolg,
    Goliath,
    Lizardfolk,
    Kenku,
    Tabaxi,
    Triton,
]

EE_races = [Aarakocra, DeepGnome, AirGenasi, FireGenasi, EarthGenasi, WaterGenasi]

MONSTER_races = [BugBear, Goblin, HobGoblin, Kobold, Orc, PureBlood]

RFTLW_races = [Kalashtar]

# Guildmaster's Guide to Ravnica
GGTR_races = [Goblin]

available_races = (
    PHB_races + VOLO_races + EE_races + MONSTER_races + RFTLW_races + GGTR_races
)

__all__ = tuple([r.name for r in available_races]) + (
    "available_races",
    "PHB_races",
    "VOLO_races",
    "EE_races",
    "MONSTER_races",
    "RFTLW_races",
    "GGTR_races",
)



from collections import defaultdict

from dungeonsheets import features, weapons
from dungeonsheets.classes.classes import CharClass, SubClass


# PHB
class Champion(SubClass):
    """The archetypal Champion focuses on the development of raw physical power
    honed to deadly perfection. Those who model themselves on this archetype
    combine rigorous training with physical excellence to deal devastating
    blows.

    """

    name = "Champion"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.ImprovedCritical]
    features_by_level[7] = [features.RemarkableAthlete]
    features_by_level[10] = [features.AdditionalFightingStyle]
    features_by_level[15] = [features.SuperiorCritical]
    features_by_level[18] = [features.Survivor]


class BattleMaster(SubClass):
    """Those who emulate the archetypal Battle Master employ martial techniques
    passed down through generations. To a Battle Master, combat is an academic
    field, sometimes including subjects beyond battle such as weaponsmithing
    and calligraphy. Not every fighter absorbs the lessons of history, theory,
    and artistry that are reflected in the Battle Master archetype, but those
    who do are well-rounded fighters of great skill and knowledge

    """

    name = "Battle Master"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.CombatSuperiority, features.StudentOfWar]
    features_by_level[7] = [features.KnowYourEnemy]
    features_by_level[15] = [features.Relentless]


class EldritchKnight(SubClass):
    """The archetypal Eldritch Knight combines the martial mastery common to all
    fighters with a careful study of magic. Eldritch Knights use magical
    techniques similar to those practiced by wizards. They focus their study on
    two of the eight schools of magic: abjuration and evocation. Abjuration
    spells grant an Eldritch Knight additional protection in battle, and
    evocation spells deal damage to many foes at once, extending the fighter's
    reach in combat. These knights learn a comparatively small number of
    spells, committing them to memory instead of keeping them in a spellbook.

    """

    name = "Eldritch Knight"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.EldritchKnightSpellcasting, features.WeaponBond]
    features_by_level[7] = [features.WarMagic]
    features_by_level[10] = [features.EldritchStrike]
    features_by_level[15] = [features.ArcaneCharge]
    features_by_level[18] = [features.ImprovedWarMagic]
    spellcasting_ability = "intelligence"
    spell_slots_by_level = {
        # char_lvl: (cantrips, 1st, 2nd, 3rd, ...)
        1: (0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
        2: (0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
        3: (2, 2, 0, 0, 0, 0, 0, 0, 0, 0),
        4: (2, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        5: (2, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        6: (2, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        7: (2, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        8: (2, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        9: (2, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        10: (3, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        11: (3, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        12: (3, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        13: (3, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        14: (3, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        15: (3, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        16: (3, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        17: (3, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        18: (3, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        19: (3, 4, 3, 3, 1, 0, 0, 0, 0, 0),
        20: (3, 4, 3, 3, 1, 0, 0, 0, 0, 0),
    }


# SCAG
class PurpleDragonKnight(SubClass):
    """Purple Dragon knights are warriors who hail from the kingdom of
    Cormyr. Pledged to protect the crown, they take the fight against evil
    beyond their kingdom's borders. They are tasked with wandering the land as
    knights errant, relying on their judgment, bravery, and fidelity to the
    code of chivalry to guide them in defeating evildoers.

    A Purple Dragon knight inspires greatness in others by committing brave
    deeds in battle. The mere presence of a knight in a hamlet is enough to
    cause some ores and bandits to seek easier prey. A lone knight is a skilled
    warrior, but a knight leading a band of allies can transform even the most
    poorly equipped militia into a ferocious war band.

    A knight prefers to lead through deeds, not words. As a knight spearheads
    an attack, the knight's actions can awaken reserves of courage and
    conviction in allies that they never suspected they had.

    """

    name = "Purple Dragon Knight"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.RallyingCry]
    features_by_level[7] = [features.RoyalEnvoy]
    features_by_level[10] = [features.InspiringSurge]
    features_by_level[15] = [features.Bulwark]


# XGTE
class ArcaneArcher(SubClass):
    """An Arcane Archer studies a unique elven method of archery that weaves magic
    into attacks to produce supernatural effects. Arcane Archers are some of
    the most elite warriors among the elves. They stand watch over the fringes
    of elven domains, keeping a keen eye out for trespassers and using
    magic-infused arrows to defeat monsters and invaders before they can reach
    elven set- tlements. Over the centuries, the methods of these elf archers
    have been learned by members of other races who can also balance arcane
    aptitude with archery.

    """

    name = "Arcane Archer"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.ArcaneArcherLore, features.ArcaneShot]
    features_by_level[7] = [features.MagicArrow, features.CurvingShot]
    features_by_level[15] = [features.EverReadyShot]


class Cavalier(SubClass):
    """The archetypal Cavalier excels at mounted combat. Usually born among the
    nobility and raised at court, a Cavalier is equally at home leading a
    cavalry charge or exchanging repartee at a state dinner. Cavaliers also
    learn how to guard those in their charge from harm, often serving as the
    protectors of their superiors and of the weak. Compelled to right wrongs or
    earn prestige, many of these fighters leave their lives of comfort to
    embark on glorious adventure

    """

    name = "Cavalier"
    features_by_level = defaultdict(list)
    features_by_level[3] = [
        features.BonusProficiencyCavalier,
        features.BornToTheSaddle,
        features.UnwaveringMark,
    ]
    features_by_level[7] = [features.WardingManeuver]
    features_by_level[10] = [features.HoldTheLine]
    features_by_level[15] = [features.FerociousCharger]
    features_by_level[18] = [features.VigilantDefender]


class Samurai(SubClass):
    """The Samurai is a fighter who draws on an implacable fighting spirit to
    overcome enemies. A Samurai's resolve is nearly unbreakable, and the
    enemies in a Samurai's path have two choices: yield or die fighting

    """

    name = "Samurai"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.BonusProficiencySamurai, features.FightingSpirit]
    features_by_level[7] = [features.ElegantCourtier]
    features_by_level[10] = [features.TirelessSpirit]
    features_by_level[15] = [features.RapidStrike]
    features_by_level[18] = [features.StrengthBeforeDeath]


# Custom
class Gunslinger(SubClass):
    """Most warriors and combat specialists spend their years perfecting the
    classic arts of swordplay, archery, or polearm tactics. Whether duelist or
    infantry, martial weapons were seemingly perfected long ago, and the true
    challenge is to master them.

    However, some minds couldn't stop with the innovation of the
    crossbow. Experimentation with alchemical components and rare metals have
    unlocked the secrets of controlled explosive force. The few who survive
    these trials of ingenuity may become the first to create, and deftly wield,
    the first firearms.

    This archetype focuses on the ability to design, craft, and utilize
    powerful, yet dangerous ranged weapons. Through creative innovation and
    immaculate aim, you become a distance force of death on the
    battlefield. However, not being a perfect science, firearms carry an
    inherent instability that can occastionally leave you without a functional
    means of attack. This is the danger of new, untested technologies in a
    world where arcane energies that rule the elements are ever present.

    Should this path of powder, fire, and metal call to you, keep your wits
    about you, hold on to your convictions as a fighter, and let skill meet
    luck to guide your bullets to strike true.

    """

    name = "Gunslinger"
    weapon_proficiencies = weapons.firearms
    _proficiencies_text = ("firearms", "tinker's tools")
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.Gunsmith, features.AdeptMarksman]
    features_by_level[7] = [features.QuickDraw]
    features_by_level[10] = [features.RapidRepair]
    features_by_level[15] = [features.LightningReload]
    features_by_level[18] = [features.ViciousIntent, features.HemorrhagingCritical]


class Fighter(CharClass):
    name = "Fighter"
    hit_dice_faces = 10
    subclass_select_level = 3
    saving_throw_proficiencies = ("strength", "constitution")
    primary_abilities = (
        "strength",
        "dexterity",
    )
    _proficiencies_text = ("All armor", "shields", "simple weapons", "martial weapons")
    weapon_proficiencies = (weapons.SimpleWeapon, weapons.MartialWeapon)
    multiclass_weapon_proficiencies = weapon_proficiencies
    _multiclass_proficiencies_text = (
        "light armor",
        "medium armor",
        "shields",
        "simple weapons",
        "martial weapons",
    )
    class_skill_choices = (
        "Acrobatics",
        "Animal Handling",
        "Athletics",
        "History",
        "Insight",
        "Intimidation",
        "Perception",
        "Survival",
    )
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.FighterFightingStyle, features.SecondWind]
    features_by_level[2] = [features.ActionSurge]
    features_by_level[5] = [features.ExtraAttackFighter]
    features_by_level[9] = [features.Indomitable]
    subclasses_available = (
        Champion,
        BattleMaster,
        EldritchKnight,
        PurpleDragonKnight,
        ArcaneArcher,
        Cavalier,
        Samurai,
        Gunslinger,
    )



from dungeonsheets.spells.spells import Spell


class EarthTremor(Spell):
    """You cause a tremor in the ground in a 10-foot radius. Each creature other than
    you in that area must make a Dexterity saving throw. On a failed save, a
    creature takes 1d6 bludgeoning damage and is knocked prone. If the ground in
    that area is loose earth or stone, it becomes difficult terrain until cleared.
    At Higher Levels. When you cast this spell using a spell slot of 2nd level or
    higher, the damage increases by 1d6 for each slot level above 1st.
    """

    name = "Earth Tremor"
    level = 1
    casting_time = "1 action"
    casting_range = "Self (10-foot radius)"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Bard", "Druid", "Sorcerer", "Wizard")


class Earthbind(Spell):
    """Choose one creature you can see within range. Yellow strips of magical energy
    loop around the creature. The target must succeed on a Strength saving throw or
    its flying speed (if any) is reduced to 0 feet for the spell's duration. An
    airborne creature affected by this spell descends at 60 feet per round until it
    reaches the ground or the spell ends.
    """

    name = "Earthbind"
    level = 2
    casting_time = "1 action"
    casting_range = "300 feet"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Sorcerer", "Warlock", "Wizard")


class Earthquake(Spell):
    """You create a seismic disturbance at a point on the ground that you can see
    within range.
    For the duration, an intense tremor rips through the ground in a
    100-foot- radius circle centered on that point and shakes creatures and
    structures in contact with the ground in that area.

    The ground in the area
    becomes difficult terrain. Each creature on the ground that is concentrating
    must make a Constitution saving throw. On a failed save, the creature's
    concentration is broken.

    When you cast this spell and at the end of each turn
    you spend concentrating on it, each creature on the ground in the area must make
    a Dexterity saving throw. On a failed save, the creature is knocked prone.


    This spell can have additional effects depending on the terrain in the area, as
    determined by the DM.
    Fissures.
    Fissures open throughout the spell's area at
    the start of your next turn after you cast the spell. A total of 1d6 such
    fissures open in locations chosen by the DM. Each is 1d10 x 10 feet deep, 10
    feet wide, and extends from one edge of the spell's area to the opposite side. A
    creature standing on a spot where a fissure opens must succeed on a Dexterity
    saving throw or fall in. A creature that successfully saves moves with the
    fissure's edge as it opens.
    A fissure that opens beneath a structure causes it
    to automatically collapse (see below).

    Structures.
    The tremor deals 50
    bludgeoning damage to any structure in contact with the ground in the area when
    you cast the spell and at the start of each of your turns until the spell ends.
    If a structure drops to 0 hit points, it collapses and potentially damages
    nearby creatures. A creature within half the distance of a structure's height
    must make a Dexterity saving throw. On a failed save, the creature takes 5d6
    bludgeoning damage, is knocked prone, and is buried in the rubble, requiring a
    DC 20 Strength (Athletics) check as an action to escape. The DM can adjust the
    DC higher or lower, depending on the nature of the rubble. On a successful save,
    the creature takes half as much damage and doesn't fall prone or become buried.
    """

    name = "Earthquake"
    level = 8
    casting_time = "1 action"
    casting_range = "500 feet"
    components = ("V", "S", "M")
    materials = "A pinch o f dirt, a piece o f rock, and a lump of clay"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ("Cleric", "Druid", "Sorcerer")


class EldritchBlast(Spell):
    """A beam of crackling energy streaks toward a creature within range. Make a ranged
    spell attack against the target. On a hit, the target takes 1d10 force damage.


    **At Higher Levels:** The spell creates more than one beam when you reach higher
    levels:
    Two beams at 5th level
    Three beams at 11th level
    Four beams at 17th
    level.
    You can direct the beams at the same target or at different ones. Make
    a separate attack roll for each beam.
    """

    name = "Eldritch Blast"
    level = 0
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Warlock",)


class ElementalBane(Spell):
    """Choose one creature you can see within range, and choose one of the following
    damage types: acid, cold, fire, lightning, or thunder.
    The target must succeed
    on a Constitution saving throw or be affected by the spell for its duration. The
    first time each turn the affected target takes damage of the chosen type, the
    target takes an extra 2d6 damage of that type. Moreover, the target loses any
    resistance to that damage type until the spell ends.

    **At Higher Levels:** When you
    cast this spell using a spell slot of 5th level or higher, you can target one
    additional creature for each slot level above 4th. The creatures must be within
    30 feet of each other when you target them.
    """

    name = "Elemental Bane"
    level = 4
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Wizard", "Warlock")


class ElementalWeapon(Spell):
    """A nonmagical weapon you touch becomes a magic weapon.
    Choose one of the
    following damage types: acid, cold, fire, lightning, or thunder. For the
    duration, the weapon has a +1 bonus to attack rolls and deals an extra 1d4
    damage of the chosen type when it hits.

    **At Higher Levels:** When you cast this
    spell using a spell slot of 5th or 6th level, the bonus to attack rolls
    increases to +2 and the extra damage increases to 2d4.
    When you use a spell
    slot of 7th level or higher, the bonus increases to +3 and the extra damage
    increases to 3d4.
    """

    name = "Elemental Weapon"
    level = 3
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Paladin",)


class EnemiesAbound(Spell):
    """You reach into the mind of one creature you can see and force it to make an
    Intelligence saving throw. A creature automatically succeeds if it is immune to
    being frightened. On a failed save, the target loses the ability to distinguish
    friend from foe, regarding all creatures it can see as enemies until the spell
    ends. Each time the target takes damage, it can repeat the saving throw, ending
    the effect on itself on a success. Whenever the affected creature chooses
    another creature as a target, it must choose the target at random from among the
    creatures it can see within range of the attack, spell, or other ability it's
    using. If an enemy provokes an opportunity attack from the affected creature,
    the creature must make that attack if it is able to.
    """

    name = "Enemies Abound"
    level = 3
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class Enervation(Spell):
    """A tendril of inky darkness reaches out from you, touching a creature you can see
    within range to drain life from it. The target must make a Dexterity saving
    throw. On a successful save, the target takes 2d8 necrotic damage, and the spell
    ends. On a failed save, the target takes 4d8 necrotic damage, and until the
    spell ends, you can use your action on each of your turns to automatically deal
    4d8 necrotic damage to the target. The spell ends ifyou use your action to do
    anything else, if the target is ever outside the spell's range, or if the target
    has total cover from you. Whenever the spell deals damage to a target, you
    regain hit points equal to half the amount of necrotic damage the target takes.


    **At Higher Levels:** When you cast this spell using a spell slot of 6th level or
    higher, the damage increases by 1d8 for each slot level above 5th.
    """

    name = "Enervation"
    level = 5
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Sorcerer", "Warlock", "Wizard")


class EnhanceAbility(Spell):
    """You touch a creature and bestow upon it a magical enhancement. Choose one of the
    following effects: the target gains the effect until the spell ends.
    - Bear's
    Endurance. The target has advantage on Constitution checks. It also gains 2d6
    temporary hit points, which are lost when the spell ends.
    - Bull's Strength. The
    target has advantage on Strength checks, and his or her carrying capacity
    doubles.
    - Cat's Grace. The target has advantage on Dexterity checks. It also
    doesn't take damage from falling 20 feet or less if it isn't incapacitated.
    -
    Eagle's Splendor. The target has advantage on Charisma checks.
    - Fox's Cunning.
    The target thas advantage on Intelligence checks.
    - Owl's Wisdom. The target has
    advantage on Wisdom checks.

    **At Higher Levels:** When you cast this spell using a
    spell slot of 3rd level or higher, you can target one additional creature for
    each slot level above 2nd.
    """

    name = "Enhance Ability"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "Fur or a feather from a beast"
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Bard", "Cleric", "Druid", "Sorcerer")


class Enlargereduce(Spell):
    """You cause a creature or an object you can see within range to grow larger or
    smaller for the duration. Choose either a creature or an object that is neither
    worn nor carried. If the target is unwilling, it can make a Constitution saving
    throw. On a success, the spell has no effect.

    If the target is a creature,
    everything it is wearing and carrying changes size with it. Any item dropped by
    an affected creature returns to normal size at once.

    Enlarge
    The target's
    size doubles in all dimensions, and its weight is multiplied by eight. This
    growth increases its size by one category – from Medium to Large, for example.
    If there isn't enough room for the target to double its size, the creature or
    object attains the maximum possible size in the space available. Until the spell
    ends, the target also has advantage on Strength checks and Strength saving
    throws. The target's weapons also grow to match its new size. While these
    weapons are enlarged, the target's attack with them deal 1d4 extra damage.


    Reduce
    The target's size is halved in all dimensions, and its weight is reduced
    to one-eighth of normal. This reduction decreases its size by one category –
    from Medium to Small, for example. Until the spell ends, the target also has
    disadvantage on Strength checks and Strength saving throws. The target's weapons
    also shrink to match its new size. While these weapons are reduced, the
    target's attacks with them deal 1d4 less damage (this can't reduce the damage
    below 1).
    """

    name = "Enlarge/Reduce"
    level = 2
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S", "M")
    materials = "A pinch of powdered iron"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Sorcerer", "Wizard")


class EnsnaringStrike(Spell):
    """The next time you hit a creature with a weapon attack before this spell ends, a
    writhing mass of thorny vines appears at the point of impact, and the target
    must succeed on a Strength saving throw or be restrained by the magical vines
    until the spell ends. A Large or larger creature has advantage on this saving
    throw. If the target succeeds on the save, the vines shrivel away.

    While
    restrained by this spell, the target takes 1d6 piercing damage at the start of
    each of its turns. A creature restrained by the vines or one that can touch the
    creature can use its action to make a Strength check against your spell save DC.
    On a success, the target is freed.

    **At Higher Levels:** If you cast this spell
    using a spell slot of 2nd level or higher, the damage increases by 1d6 for each
    slot level above 1st.
    """

    name = "Ensnaring Strike"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Ranger",)


class Entangle(Spell):
    """Grasping weeds and vines sprout from the ground in a 20-foot square starting
    from a point within range. For the duration, these plants turn the ground in the
    area into difficult terrain.

    A creature in the area when you cast the spell
    must succeed on a Strength saving throw or be restrained by the entangling
    plants until the spell ends. A creature restrained by the plants can use its
    action to make a Strength check against your spell save DC. On a success, it
    frees itself.

    When the spell ends, the conjured plants wilt away.
    """

    name = "Entangle"
    level = 1
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid",)


class Enthrall(Spell):
    """You weave a distracting string of words, causing creatures of your choice that
    you can see within range and that can hear you to make a Wisdom saving throw.
    Any creature that can't be charmed succeeds on this saving throw automatically,
    and if you or your companions are fighting a creature, it has advantage on the
    save. On a failed save, the target has disadvantage on Wisdom (Perception)
    checks made to perceive any creature other than you until the spell ends or
    until the target can no longer hear you. The spell ends if you are incapacitated
    or can no longer speak.
    """

    name = "Enthrall"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Warlock")


class EruptingEarth(Spell):
    """Choose a point you can see on the ground within range. A fountain of churned
    earth and stone erupts in a 20-foot cube centered on that point. Each creature
    in that area must make a Dexterity saving throw. A creature takes 3d12
    bludgeoning damage on a failed save, or half as much damage on a successful one.
    Additionally, the ground in that area becomes difficult terrain until cleared
    away. Each 5-foot-square portion of the area requires at least 1 minute to clear
    by hand.

    **At Higher Levels:** When you cast this spell using a spell slot of 4rd
    level or higher, the damage increases by 1d12 for each slot level above 3rd.
    """

    name = "Erupting Earth"
    level = 3
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S", "M")
    materials = "A piece of obsidian"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Sorcerer", "Wizard")


class Etherealness(Spell):
    """You step into the border regions of the Ethereal Plane, in the area where it
    overlaps with your current plane. You remain in the Border Ethereal for the
    duration or until you use your action to dismiss the spell. During this time,
    you can move in any direction. If you move up or down, every foot of movement
    costs an extra foot. You can see and hear the plan you originated from, but
    everything there looks gray, and you can't see anything more than 60 feet away.


    While on the Ethereal Plane, you can only affect and be affected by other
    creatures on that plane. Creatures that aren't on the Ethereal Plance can't
    perceive you and can't interact with you, unless a special ability or magic has
    given them the ability to do so.

    You ignore all objects and effects that
    aren't on the Ethereal Plane, allowing you to move through objects you perceive
    on the plan you originated from. When the spell ends, you immediately return to
    the plane you originiated from in teh spot you currently occupy. If you occupy
    the same spot as a solid object or creature when this happens, you are
    imediately shunted to the neares unoccupied space that you can occupy and take
    force damage equal to twice the number of feet you are moved.

    This spell has
    no effect if you cast it while you are on the Ethereal Plane or a plane that
    doesn't border it, such as one of the Outer Planes.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 8th level or higher, you can target up to three willing
    creatures (including you) for each slot level above 7th. The
    creatures must be within 10 feet of you when you cast the spell.

    """

    name = "Etherealness"
    level = 7
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Up to 8 hours"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Bard", "Cleric", "Sorcerer", "Warlock", "Wizard")


class EvardsBlackTentacles(Spell):
    """Squirming, ebony tentacles fill a 20-foot square on ground that you can see
    within range. For the duration, these tentacles turn the ground in the area into
    difficult terrain.

    When a creature enters the affected area for the first
    time on a turn or starts its turn there, the creature must succeed on a
    Dexterity saving throw or take 3d6 bludgeoning damage and be restrained by the
    tentacles until the spell ends. A creature that starts its turn in the area and
    is already restrained by the tentacles takes 3d6 bludgeoning damage.

    A
    creature restrained by the tentacles can use its action to make a Strength or
    Dexterity check (its choice) against your spell save DC. On a success, it frees
    itself.
    """

    name = "Evards Black Tentacles"
    level = 4
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ("V", "S", "M")
    materials = "A piece of tentacle from a giant octopus or a giant squid"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Wizard",)


class ExpeditiousRetreat(Spell):
    """This spell allows you to move at an incredible pace. When you cast this spell,
    and then as a bonus action on each of your turns until the spell ends, you can
    take the Dash action.
    """

    name = "Expeditious Retreat"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Sorcerer", "Warlock", "Wizard")


class Eyebite(Spell):
    """For the spell's duration, your eyes become an inky void imbued with
    dread power.  One creature of your choice within 60 feet of you
    that you can see must succeed on a Wisdom saving throw or be
    affected by one of the following effects of your choice for the
    duration. On each of your turns until the spell ends, you can use
    your action to target another creature but can't target a creature
    again if it has succeeded on a saving throw against this casting
    of eyebite.

    Asleep
      The target falls unconscious. It wakes up if it takes any damage
      or if another creature uses its action to shake the sleeper
      awake.

    Panicked
      The target is frightened of you. On each of its turns, the
      frightened creature must take the Dash action and move away from
      you by the safest and shortest available route, unless there is
      nowhere to move. If the target moves to a place at least 60 feet
      away from you where it can no longer see you, this effect ends.

    Sickened
      The target has disadvantage on attack rolls and ability
      checks. At the end of each of its turns, it can make another
      Wisdom saving throw. If it succeeds, the effect ends.

    """

    name = "Eyebite"
    level = 6
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")



"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability


class IceDevil(Monster):
    """
    Devil's Sight.
      Magical darkness doesn't impede the devil's darkvision.
    Magic Resistance.
      The devil has advantage on saving throws against spells and other
      magical effects.
    Multiattack.
      The devil makes three attacks: one with its bite, one with its claws,
      and one with its tail.
    Bite.
      Melee Weapon Attack: +10 to hit, reach 5 ft., one target. Hit: 12 (2d6
      + 5) piercing damage plus 10 (3d6) cold damage.
    Claws.
      Melee Weapon Attack: +10 to hit, reach 5 ft., one target. Hit: 10 (2d4
      + 5) slashing damage plus 10 (3d6) cold damage.
    Tail.
      Melee Weapon Attack:+10 to hit, reach 10 ft., one target. Hit: 12 (2d6
      + 5) bludgeoning damage plus 10 (3d6) cold damage.
    Wall of Ice.
      The devil magically forms an opaque wall of ice on a solid surface it
      can see within 60 feet of it. The wall is 1 foot thick and up to 30
      feet long and 10 feet high, or it's a hemispherical dome up to 20 feet
      in diameter.
      
      When the wall appears, each creature in its space is pushed out of it
      by the shortest route. The creature chooses which side of the wall to
      end up on, unless the creature is incapacitated. The creature then
      makes a DC 17 Dexterity saving throw, taking 35 (10d6) cold damage on
      a failed save, or half as much damage on a successful one.
      
      The wall lasts for 1 minute or until the devil is incapacitated or
      dies. The wall can be damaged and breached; each 10-foot section has
      AC 5, 30 hit points, vulnerability to fire damage, and immunity to
      acid, cold, necrotic, poison, and psychic damage. If a section is
      destroyed, it leaves behind a sheet of frigid air in the space the
      wall occupied. Whenever a creature finishes moving through the frigid
      air on a turn, willingly or otherwise, the creature must make a DC 17
      Constitution saving throw, taking 17 (5d6) cold damage on a failed
      save, or half as much damage on a successful one. The frigid air
      dissipates when the rest of the wall vanishes.
    """
    name = "Ice Devil"
    description = "Large fiend, lawful evil"
    challenge_rating = 14
    armor_class = 18
    skills = ""
    senses = "Blindsight 60 ft., Darkvision 120 ft., Passive Perception 12"
    languages = "Infernal, telepathy 120 ft."
    strength = Ability(21)
    dexterity = Ability(14)
    constitution = Ability(18)
    intelligence = Ability(18)
    wisdom = Ability(15)
    charisma = Ability(18)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 180
    hit_dice = "19d10"
    spells = []


class IceMephit(Monster):
    """
    Death Burst.
      When the mephit dies, it explodes in a burst of jagged ice. Each
      creature within 5 ft. of it must make a DC 10 Dexterity saving throw,
      taking 4 (1d8) slashing damage on a failed save, or half as much
      damage on a successful one.
    False Appearance.
      While the mephit remains motionless, it is indistinguishable from an
      ordinary shard of ice.
    Innate Spellcasting.
      The mephit can innately cast fog cloud, requiring no material
      components. Its innate spellcasting ability is Charisma.
    Claws.
      Melee Weapon Attack: +3 to hit, reach 5 ft., one creature. Hit: 3 (1d4
      + 1) slashing damage plus 2 (1d4) cold damage.
    Frost Breath (Recharge 6).
      The mephit exhales a 15-foot cone of cold air. Each creature in that
      area must succeed on a DC 10 Dexterity saving throw, taking 5 (2d4)
      cold damage on a failed save, or half as much damage on a successful
      one.
    Variant: Summon Mephits.
      The mephit has a 25 percent chance of summoning 1d4 mephits of its
      kind. A summoned mephit appears in an unoccupied space within 60 feet
      of its summoner, acts as an ally of its summoner, and can't summon
      other mephits. It remains for 1 minute, until it or its summoner dies,
      or until its summoner dismisses it as an action.
    """
    name = "Ice Mephit"
    description = "Small elemental, neutral evil"
    challenge_rating = 0.5
    armor_class = 11
    skills = "Perception +2, Stealth +3"
    senses = "Darkvision 60 ft., Passive Perception 12"
    languages = "Aquan, Auran"
    strength = Ability(7)
    dexterity = Ability(13)
    constitution = Ability(10)
    intelligence = Ability(9)
    wisdom = Ability(11)
    charisma = Ability(12)
    speed = 30
    swim_speed = 0
    fly_speed = 30
    climb_speed = 0
    hp_max = 21
    hit_dice = "6d6"
    spells = []


class Imp(Monster):
    """
    Shapechanger.
      The imp can use its action to polymorph into a beast form that
      resembles a rat (speed 20 ft.), a raven (20 ft., fly 60 ft.), or a
      spider (20 ft., climb 20 ft.), or back into its true form. Its
      statistics are the same in each form, except for the speed changes
      noted. Any equipment it is wearing or carrying isn't transformed. It
      reverts to its true form if it dies.
    Devil's Sight.
      Magical darkness doesn't impede the imp's darkvision.
    Magic Resistance.
      The imp has advantage on saving throws against spells and other
      magical effects.
    Variant: Familiar.
      The imp can serve another creature as a familiar, forming a telepathic
      bond with its willing master. While the two are bonded, the master can
      sense what the quasit senses as long as they are within 1 mile of each
      other. While the imp is within 10 feet of its master, the master
      shares the quasit's Magic Resistance trait. At any time and for any
      reason, the imp can end its service as a familiar, ending the
      telepathic bond.
    Sting (Bite in Beast Form).
      Melee Weapon Attack: +5 to hit, reach 5 ft ., one target. Hit: 5 (1d4
      + 3) piercing damage, and the target must make on a DC 11 Constitution
      saving throw, taking 10 (3d6) poison damage on a failed save, or half
      as much damage on a successful one.
    Invisibility.
      The imp magically turns invisible until it attacks, or until its
      concentration ends (as if concentrating on a spell). Any equipment the
      imp wears or carries is invisible with it.
    """
    name = "Imp"
    description = "Tiny fiend, lawful evil"
    challenge_rating = 1
    armor_class = 13
    skills = "Deception +4, Insight +3, Persuasion +4, Stealth +5"
    senses = "Darkvision 120 ft., Passive Perception 11"
    languages = "Infernal, Common"
    strength = Ability(6)
    dexterity = Ability(17)
    constitution = Ability(13)
    intelligence = Ability(11)
    wisdom = Ability(12)
    charisma = Ability(14)
    speed = 20
    swim_speed = 0
    fly_speed = 40
    climb_speed = 0
    hp_max = 10
    hit_dice = "3d4"
    spells = []


class InvisibleStalker(Monster):
    """
    Invisibility.
      The stalker is invisible.
    Faultless Tracker.
      The stalker is given a quarry by its summoner. The stalker knows the
      direction and distance to its quarry as long as the two of them are on
      the same plane of existence. The stalker also knows the location of
      its summoner.
    Multiattack.
      The stalker makes two slam attacks.
    Slam.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 10 (2d6
      + 3) bludgeoning damage.
    """
    name = "Invisible Stalker"
    description = "Medium elemental, neutral"
    challenge_rating = 6
    armor_class = 14
    skills = "Perception +8, Stealth +10"
    senses = "Darkvision 60 ft., Passive Perception 18"
    languages = "Auran, understands Common but doesn't speak it"
    strength = Ability(16)
    dexterity = Ability(19)
    constitution = Ability(14)
    intelligence = Ability(10)
    wisdom = Ability(15)
    charisma = Ability(11)
    speed = 50
    swim_speed = 0
    fly_speed = 50
    climb_speed = 0
    hp_max = 104
    hit_dice = "16d8"
    spells = []


class IronGolem(Monster):
    """
    Fire Absorption.
      Whenever the golem is subjected to fire damage, it takes no damage and
      instead regains a number of hit points equal to the fire damage dealt.
    Immutable Form.
      The golem is immune to any spell or effect that would alter its form.
    Magic Resistance.
      The golem has advantage on saving throws against spells and other
      magical effects.
    Magic Weapons.
      The golem's weapon attacks are magical.
    Multiattack.
      The golem makes two melee attacks.
    Slam.
      Melee Weapon Attack: +13 to hit, reach 5 ft., one target. Hit: 20 (3d8
      + 7) bludgeoning damage.
    Sword.
      Melee Weapon Attack: +13 to hit, reach 10 ft., one target. Hit: 23
      (3d10 + 7) slashing damage.
    Poison Breath.
      The golem exhales poisonous gas in a 15-foot cone. Each creature in
      that area must make a DC 19 Constitution saving throw, taking 45
      (l0d8) poison damage on a failed save, or half as much damage on a
      successful one.
    """
    name = "Iron Golem"
    description = "Large construct, unaligned"
    challenge_rating = 16
    armor_class = 20
    skills = ""
    senses = "Darkvision 120 ft., Passive Perception 10"
    languages = "understands the languages of its creator but can't speak"
    strength = Ability(24)
    dexterity = Ability(9)
    constitution = Ability(20)
    intelligence = Ability(3)
    wisdom = Ability(11)
    charisma = Ability(1)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 210
    hit_dice = "20d10"
    spells = []



from dungeonsheets.spells.spells import Spell, create_spell, all_spells

from dungeonsheets.spells.spells_a import *
from dungeonsheets.spells.spells_b import *
from dungeonsheets.spells.spells_c import *
from dungeonsheets.spells.spells_d import *
from dungeonsheets.spells.spells_e import *
from dungeonsheets.spells.spells_f import *
from dungeonsheets.spells.spells_g import *
from dungeonsheets.spells.spells_h import *
from dungeonsheets.spells.spells_i import *
from dungeonsheets.spells.spells_j import *
from dungeonsheets.spells.spells_k import *
from dungeonsheets.spells.spells_l import *
from dungeonsheets.spells.spells_m import *
from dungeonsheets.spells.spells_n import *
from dungeonsheets.spells.spells_o import *
from dungeonsheets.spells.spells_p import *
from dungeonsheets.spells.spells_q import *
from dungeonsheets.spells.spells_r import *
from dungeonsheets.spells.spells_s import *
from dungeonsheets.spells.spells_t import *
from dungeonsheets.spells.spells_u import *
from dungeonsheets.spells.spells_v import *
from dungeonsheets.spells.spells_w import *
from dungeonsheets.spells.spells_x import *
from dungeonsheets.spells.spells_y import *
from dungeonsheets.spells.spells_z import *

from dungeonsheets.content_registry import default_content_registry


default_content_registry.add_module(__name__)



"""A collection of monsters. Also useful for building a list of wild
shape forms.

"""

from abc import ABCMeta


from dungeonsheets.content import Creature
from dungeonsheets.spells import Spell


xp_by_challenge_rating = {
    0: 10,
    (1 / 8): 25,
    (1 / 4): 50,
    (1 / 2): 100,
    1: 200,
    2: 450,
    3: 700,
    4: 1100,
    5: 1800,
    6: 2300,
    7: 2900,
    8: 3900,
    9: 5000,
    10: 5900,
    11: 7200,
    12: 8400,
    13: 10000,
    14: 11500,
    15: 13000,
    16: 15000,
    17: 18000,
    18: 20000,
    19: 22000,
    20: 25000,
    21: 33000,
    22: 41000,
    23: 50000,
    24: 62000,
    25: 75000,
    26: 90000,
    27: 105000,
    28: 120000,
    29: 130000,
    30: 155000,
}


def challenge_rating_to_xp(cr):
    """Determine the XP awarded for slaying a monster with the given
    challenge rating *cr*.

    """
    xp = xp_by_challenge_rating[cr]
    return "{:,}".format(xp)


class SpellFactory(ABCMeta):
    """Meta class to resolve spell strings into the ``spells.Spell``.
    
    For classes using this metaclass, the *spell* attribute, if
    present, should be a list of spells that the entity knows. For
    each entry on that list, anything that is not already a spell
    class (so probably a string) will be resolved into the
    corresponding ``spells.Spell`` class.
    
    """
    def __init__(self, name, bases, attrs):
        for idx, spell in enumerate(self.spells):
            TheSpell = self._resolve_mechanic(spell, SuperClass=Spell)
            self.spells[idx] = TheSpell


class Monster(Creature, metaclass=SpellFactory):

    """A monster that may be encountered when adventuring."""

    name = "Generic Monster"
    description = ""
    challenge_rating = 0
    skills = "Perception +3, Stealth +4"
    damage_resistances = ""
    damage_immunities = ""
    damage_vulnerabilities = ""
    condition_immunities = ""
    saving_throws = ""
    # TODO: Consider refactoring stats.Speed to consider all of these
    # just like we do stats.Ability

    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    burrow_speed = 0
    hp_max = 10
    hit_dice = "1d6"
    spells = []

    def __init__(self):
        super(Monster, self).__init__()

    @property
    def is_beast(self):
        is_beast = "beast" in self.description.lower()
        return is_beast



from dungeonsheets.content_registry import default_content_registry


default_content_registry.add_module(__name__)


class Infusion:
    name = "Unknown infusion"
    item = "Item to be infused"
    prerequisite = ""
    classes = ("Artificer",)

    def __str__(self):
        indicator = ("$", self.special_material)
        if indicator:
            return self.name + f' ({"".join(indicator)})'
        else:
            return self.name

    def __repr__(self):
        return '"{:s}"'.format(self.name)

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return 0

    @property
    def special_material(self):
        return "worth at least" in self.item.lower()


class BootsOfTheWindingPath(Infusion):
    """While wearing these boots, a creature can teleport up to 15 feet as a
    bonus action to an unoccupied space the creature can see. The creature
    must have occupied that space at some point during the current turn.
    """

    name = "Boots of the Winding Path"
    item = "A pair of boots (requires attunement"
    prerequisite = "6th-level artificer"


class EnhancedArcaneFocus(Infusion):
    """While holding this item, a creature gains a + 1 bonus to spell attack
    rolls. In addition, the creature ignores half cover when making a spell
    attack.

    The bonus increases to +2 when you reach 10th level in this class.
    """

    name = "Enhanced Arcane Focus"
    item = "A rod, staff, or wand (requires attunement)"


class EnhancedDefense(Infusion):
    """A creature gains a + 1 bonus to Armor Class while wearing (armor) or
    wielding (shield) the infused item.

    The bonus increases to +2 when you reach 10th level in this class.
    """

    name = "Enhanced Defense"
    item = "A suit of armor or shield"


class EnhancedWeapon(Infusion):
    """This magic weapon grants a +1 bonus to attack and damage rolls made with
    it.

    The bonus increases to +2 when you reach 10th level in this class.
    """

    name = "Enhanced Weapon"
    item = "A simple or martial weapon"


class HomunculusServant(Infusion):
    """You learn intricate methods for magically creating a special homunculus
    that serves you. The item you infuse serves as the creature's heart, around
    which the creature's body instantly forms.

    You determine the homunculus's appearance. Some artificers prefer
    mechanical-looking birds, whereas some like winged vials or miniature,
    animate cauldrons.

    The homunculus is friendly to you and your companions, and it obeys your
    commands. See this creature's game statistics in the Homunculus Servant
    stat block.

    In combat, the homunculus shares your initiative count, but it takes its
    turn immediately after yours. It can move and use its reaction on its own,
    but the only action it takes on its turn is the Dodge action, unless you
    take a bonus action on your turn to command it to take the action in its
    stat block or the Dash, Disengage, Help, Hide, or Search action.

    The homunculus regains 2d6 hit points if the *mending* spell is cast on it.
    If it dies, it vanishes, leaving its heart in its space.
    """

    name = "Homunculus Servant"
    item = "A gem worth at least 100gp or a dragonshard"
    prerequisite = "6th-level artificer"


class RadiantWeapon(Infusion):
    """This magic weapon grants a + 1 bonus to attack and damage rolls made
    with it. While holding it, the wielder can take a bonus action to cause it
    to shed bright light in a 30-foot radius and dim light for an additional 30
    feet. The wielder can extinguish the light as a bonus action.

    The weapon has 4 charges. As a reaction immediately after being hit by an
    attack, the wielder can expend 1 charge and cause the attacker to be
    blinded until the end of the attacker's next turn, unless the attacker
    succeeds on a Constitution saving throw against your spell save DC. The
    weapon regains ld4 expended charges daily at dawn.
    """

    name = "Radiant Weapon"
    item = "A simple or martial weapon (requires attunement)"
    prerequisite = "6th-level artificer"


class RepeatingShot(Infusion):
    """This magic weapon grants a + 1 bonus to attack and damage rolls made
    with it when it's used to make a ranged attack, and it ignores the loading
    property if it has it.

    If you load no ammunition in the weapon, it produces its own, automatically
    creating one piece of magic am­munition when you make a ranged attack with
    it. The ammunition created by the weapon vanishes the instant after it hits
    or misses a target.
    """

    name = "Repeating Shot"
    item = """A simple or martial weapon with the ammunition property (requires
              attunement)"""


class ReplicateMagicItem(Infusion):
    """Using this infusion, you replicate a particular magic item. You can
    learn this infusion multiple times; each time you do so, choose a magic
    item that you can make with it, picking from the Replicable Items tables
    below. A table's title tells you the level you must be in the class to
    choose an item from the table.

    In the tables, an item's entry tells you whether the item requires
    attunement. See the item's description in the *Dungeon Master's Guide* for
    more information about it, including the type of object required for its
    making. If you have *Xanathar's Guide to Everything*, you can choose from
    among the common magic items in that book when you pick a magic item you
    can replicate with this infusion.
    """

    name = "Replicate Magic Item"


class RepulsionShield(Infusion):
    """A creature gains a + 1 bonus to Armor Class while wield­ing this shield.

    The shield has 4 charges. While holding it, the wielder can use a reaction
    immediately after being hit by a melee attack to expend 1 of the shield's
    charges and push the attacker up to 15 feet away. The shield regains ld4
    expended charges daily at dawn.
    """

    name = "Repulsion Shield"
    item = "A shield (requires attunement)"


class ResistantArmor(Infusion):
    """While wearing this armor, a creature has resistance to one of the
    following damage types, which you choose when you infuse the item: acid,
    cold, fire, force, light­ning, necrotic, poison, psychic, radiant, or
    thunder.
    """

    name = "Resistant Armor"
    item = "A suit of armor (requires attunement)"


class ReturningWeapon(Infusion):
    """This magic weapon grants a + 1 bonus to attack and damage rolls made
    with it, and it returns to the wielder's hand immediately after it is used
    to make a ranged attack.
    """

    name = "Returning Weapon"
    item = "A simple or martial weapon with the thrown property"



"""Convenience module holding base classes for the various kinds of
game mechanics."""

from dungeonsheets.content import Content
from dungeonsheets.spells import Spell
from dungeonsheets.features import Feature
from dungeonsheets.infusions import Infusion
from dungeonsheets.weapons import (
    Weapon,
    MeleeWeapon,
    RangedWeapon,
    SimpleWeapon,
    MartialWeapon,
)
from dungeonsheets.armor import Armor, Shield
from dungeonsheets.magic_items import MagicItem
from dungeonsheets.monsters import Monster
from dungeonsheets.stats import Ability



__all__ = ("Ranger", "RevisedRanger")

from collections import defaultdict

from dungeonsheets import features, spells, weapons
from dungeonsheets.classes.classes import CharClass, SubClass


# PHB
class Hunter(SubClass):
    """Emulating the Hunter archetype means accepting your place as a bulwark
    between civilization and the terrors of the wilderness. As you walk the
    Hunter's path, you learn specialized techniques for fighting the threats
    you face, from rampaging ogres and hordes of orcs to towering giants and
    terrifying dragons.

    """

    name = "Hunter"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.HuntersPrey]
    features_by_level[7] = [features.DefensiveTactics]
    features_by_level[11] = [features.MultiattackRanger]
    features_by_level[15] = [features.SuperiorHuntersDefense]


class BeastMaster(SubClass):
    """The Beast Master archetype embodies a friendship between the civilized
    races and the beasts of the world. United in focus, beast and ranger work
    as one to fight the monstrous foes that threaten civilization and the
    wilderness alike. Emulating the Beast Master archetype means committing
    yourself to this ideal, working in partnership with an animal as its
    companion and friend.

    """

    name = "Beast Master"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.RangersCompanion]
    features_by_level[7] = [features.ExceptionalTraining]
    features_by_level[11] = [features.BestialFury]
    features_by_level[15] = [features.ShareSpells]


# XGTE
class GloomStalker(SubClass):
    """Gloom Stalkers are at home in the darkest places: deep under the earth, in
    gloomy alleyways, in primeval forests, and wherever else the light
    dims. Most folk enter such places with trepidation, but a Gloom Stalker
    ventures boldly into the darkness, seeking to ambush threats before they
    can reach the broader world. Such rangers are often found in the Underdark,
    but they will go any place Where evil lurks in the shadows

    """

    name = "Gloom Stalker"
    features_by_level = defaultdict(list)
    _spells = {
        3: [spells.DisguiseSelf],
        5: [spells.RopeTrick],
        9: [spells.Fear],
        13: [spells.GreaterInvisibility],
        17: [spells.Seeming],
    }
    features_by_level[3] = [
        features.DreadAmbusher,
        features.UmbralSight,
        features.Darkvision,
    ]
    features_by_level[7] = [features.IronMind]
    features_by_level[11] = [features.StalkersFlurry]
    features_by_level[15] = [features.ShadowyDodge]

    @property
    def level(self):
        return self.owner.Ranger.level

    @property
    def spells_known(self):
        spells = []
        for lvl, sps in self._spells.items():
            if self.level >= lvl:
                spells.extend(sps)
        return spells

    @property
    def spells_prepared(self):
        return self.spells_known


class HorizonWalker(SubClass):
    """Horizon Walkers guard the world against threats that originate from other
    planes or that seek to ravage the mortal realm with otherworldly
    magic. They seek out planar portals and keep watch over them, venturing to
    the Inner Planes and the Outer Planes as needed to pursue their foes. These
    rangers are also friends to any forces in the multiverse-especially
    benevolent dragons, fey, and elementals-that work to preserve life and the
    order of the planes

    """

    name = "Horizon Walker"
    _spells = {
        3: [spells.ProtectionFromEvilAndGood],
        5: [spells.MistyStep],
        9: [spells.Haste],
        13: [spells.Banishment],
        17: [spells.TeleportationCircle],
    }
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.DetectPortal, features.PlanarWarrior]
    features_by_level[7] = [features.EtherealStep]
    features_by_level[11] = [features.DistantStrike]
    features_by_level[15] = [features.SpectralDefense]

    @property
    def level(self):
        return self.owner.Ranger.level

    @property
    def spells_known(self):
        spells = []
        for lvl, sps in self._spells.items():
            if self.level >= lvl:
                spells.extend(sps)
        return spells

    @property
    def spells_prepared(self):
        return self.spells_known


class MonsterSlayer(SubClass):
    """You have dedicated yourself to hunting down creatures of the night and
    wielders of grim magic. A Monster Slayer seeks out vampires, dragons, evil
    fey, fiends, and other magical threats. Trained in supernatural techniques
    to overcome such monsters, Slayers are experts at unearthing and defeating
    mighty, mystical foes.

    """

    name = "Monster Slayer"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.HuntersSense, features.SlayersPrey]
    features_by_level[7] = [features.SupernaturalDefense]
    features_by_level[11] = [features.MagicUsersNemesis]
    features_by_level[15] = [features.SlayersCounter]
    _spells = {
        3: [spells.ProtectionFromEvilAndGood],
        5: [spells.ZoneOfTruth],
        9: [spells.MagicCircle],
        13: [spells.Banishment],
        17: [spells.HoldMonster],
    }

    @property
    def level(self):
        return self.owner.Ranger.level

    @property
    def spells_known(self):
        spells = []
        for lvl, sps in self._spells.items():
            if self.level >= lvl:
                spells.extend(sps)
        return spells

    @property
    def spells_prepared(self):
        return self.spells_known


class Ranger(CharClass):
    name = "Ranger"
    hit_dice_faces = 10
    saving_throw_proficiencies = ("strength", "dexterity")
    primary_abilities = ("dexterity", "wisdom")
    _proficiencies_text = (
        "light armor",
        "medium armor",
        "shields",
        "simple weapons",
        "martial weapons",
    )
    weapon_proficiencies = (weapons.SimpleWeapon, weapons.MartialWeapon)
    multiclass_weapon_proficiencies = weapon_proficiencies
    _multiclass_proficiencies_text = (
        "light armor",
        "medium armor",
        "shields",
        "simple weapons",
        "martial weapons",
        "[choose one skill from Ranger list]",
    )
    class_skill_choices = (
        "Animal Handling",
        "Athletics",
        "Insight",
        "Investigation",
        "Nature",
        "Perception",
        "Stealth",
        "Survival",
    )
    num_skill_choices = 3
    features_by_level = defaultdict(list)
    features_by_level[1] = [features.FavoredEnemy, features.NaturalExplorer]
    features_by_level[2] = [features.RangerFightingStyle]
    features_by_level[3] = [features.PrimevalAwareness]
    features_by_level[5] = [features.ExtraAttackRanger]
    features_by_level[8] = [features.LandsStride]
    features_by_level[10] = [features.HideInPlainSight]
    features_by_level[14] = [features.Vanish]
    features_by_level[18] = [features.FeralSenses]
    features_by_level[20] = [features.FoeSlayer]
    subclasses_available = (
        Hunter,
        BeastMaster,
        GloomStalker,
        HorizonWalker,
        MonsterSlayer,
    )
    spellcasting_ability = "wisdom"
    spell_slots_by_level = {
        # char_lvl: (cantrips, 1st, 2nd, 3rd, ...)
        1: (0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
        2: (0, 2, 0, 0, 0, 0, 0, 0, 0, 0),
        3: (0, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        4: (0, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        5: (0, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        6: (0, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        7: (0, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        8: (0, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        9: (0, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        10: (0, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        11: (0, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        12: (0, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        13: (0, 4, 3, 3, 1, 0, 0, 0, 0, 0),
        14: (0, 4, 3, 3, 1, 0, 0, 0, 0, 0),
        15: (0, 4, 3, 3, 2, 0, 0, 0, 0, 0),
        16: (0, 4, 3, 3, 2, 0, 0, 0, 0, 0),
        17: (0, 4, 3, 3, 3, 1, 0, 0, 0, 0),
        18: (0, 4, 3, 3, 3, 1, 0, 0, 0, 0),
        19: (0, 4, 3, 3, 3, 2, 0, 0, 0, 0),
        20: (0, 4, 3, 3, 3, 2, 0, 0, 0, 0),
    }


# Revised Ranger
class BeastConclave(SubClass):
    """Many rangers are more at home in the wilds than in civilization, to the
    point where animals consider them kin. Rangers of the Beast Conclave
    develop a close bond with a beast, then further strengthen that bond
    through the use of magic.

    """

    name = "Beast Conclave"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.AnimalCompanion, features.CompanionsBond]
    features_by_level[5] = [features.CoordinatedAttack]
    features_by_level[7] = [features.BeastsDefense]
    features_by_level[11] = [features.StormOfClawsAndFangs]
    features_by_level[15] = [features.SuperiorBeastsDefense]


class HunterConclave(SubClass):
    """Some rangers seek to master weapons to better protect civilization from the
    terrors of the wilderness. Members of the Hunter Conclave learn specialized
    fighting techniques for use against the most dire threats, from rampaging
    ogres and hordes of orcs to towering giants and terrifying dragons

    """

    name = "Hunter Conclave"
    features_by_level = defaultdict(list)
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.HuntersPrey]
    features_by_level[5] = [features.ExtraAttackRanger]
    features_by_level[7] = [features.DefensiveTactics]
    features_by_level[11] = [features.MultiattackRanger]
    features_by_level[15] = [features.SuperiorHuntersDefense]


class DeepStalkerConclave(SubClass):
    """Most folk descend into the depths of the Underdark only under the most
    pressing conditions, undertaking some desperate quest or following the
    promise of vast riches. All too often, evil festers beneath the earth
    unnoticed, and rangers of the Deep Stalker Conclave strive to uncover and
    defeat such threats before they can reach the surface.

    """

    name = "Deep Stalker Conclave"
    features_by_level = defaultdict(list)
    features_by_level[3] = [features.UnderdarkScout]
    features_by_level[5] = [features.ExtraAttackRanger]
    features_by_level[7] = [features.IronMind]
    features_by_level[11] = [features.StalkersFlurry]
    features_by_level[15] = [features.StalkersDodge]
    _spells = {
        3: [spells.DisguiseSelf],
        5: [spells.RopeTrick],
        9: [spells.GlyphOfWarding],
        13: [spells.GreaterInvisibility],
        17: [spells.Seeming],
    }

    @property
    def spells_prepared(self):
        level = self.owner.Ranger.level
        my_spells = []
        for lvl, sps in self._spells.items():
            if level >= lvl:
                my_spells.extend(sps)
        return my_spells

    @property
    def spells_known(self):
        return self.spells_prepared


class RevisedRanger(Ranger):
    name = "Revised Ranger"
    features_by_level = defaultdict(list)
    features_by_level[1] = [
        features.FavoredEnemyRevised,
        features.NaturalExplorerRevised,
    ]
    features_by_level[2] = [features.RangerFightingStyle]
    features_by_level[3] = [features.PrimevalAwarenessRevised]
    features_by_level[6] = [features.GreaterFavoredEnemy]
    features_by_level[8] = [features.FleetOfFoot]
    features_by_level[10] = [features.HideInPlainSightRevised]
    features_by_level[14] = [features.Vanish]
    features_by_level[18] = [features.FeralSenses]
    features_by_level[20] = [features.FoeSlayer]
    subclasses_available = (BeastConclave, HunterConclave, DeepStalkerConclave)



"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability


class Mage(Monster):
    """
    Spellcasting.
      The mage is a 9th-level spellcaster. Its spellcasting ability is
      Intelligence (spell save DC 14, +6 to hit with spell attacks). The
      mage has the following wizard spells prepared:
      
  
      
      - Cantrips (at will): fire bolt, light, mage hand, prestidigitation
      
      - 1st level (4 slots): detect magic, mage armor, magic missile, shield
      
      - 2nd level (3 slots): misty step, suggestion
      
      - 3rd level (3 slots): counterspell, fireball, fly
      
      - 4th level (3 slots): greater invisibility, ice storm
      
      - 5th level (1 slot): cone of cold
    Dagger.
      Melee or Ranged Weapon Attack: +5 to hit, reach 5 ft. or range 20/60
      ft., one target. Hit: 4 (1d4 + 2) piercing damage.
    """
    name = "Mage"
    description = "Medium humanoid, any alignment"
    challenge_rating = 6
    armor_class = 12
    skills = "Arcana +6, History +6"
    senses = "Passive Perception 11"
    languages = "any four languages"
    strength = Ability(9)
    dexterity = Ability(14)
    constitution = Ability(11)
    intelligence = Ability(17)
    wisdom = Ability(12)
    charisma = Ability(11)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 40
    hit_dice = "9d8"
    spells = ["fire bolt", "light", "mage hand", "prestidigitation", "detect magic", "mage armor", "magic missile", "shield", "misty step", "suggestion", "counterspell", "fireball", "fly", "greater invisibility", "ice storm", "cone of cold"]


class MagmaMephit(Monster):
    """
    Death Burst.
      When the mephit dies, it explodes in a burst of lava. Each creature
      within 5 ft. of it must make a DC 11 Dexterity saving throw, taking 7
      (2d6) fire damage on a failed save, or half as much damage on a
      successful one.
    False Appearance.
      While the mephit remains motionless, it is indistinguishable from an
      ordinary mound of magma.
    Innate Spellcasting.
      The mephit can innately cast heat metal (spell save DC 10), requiring
      no material components. Its innate spellcasting ability is Charisma.
    Claws.
      Melee Weapon Attack: +3 to hit, reach 5 ft ., one creature. Hit: 3
      (1d4 + 1) slashing damage plus 2 (1d4) fire damage.
    Fire Breath (Recharge 6).
      The mephit exhales a 15-foot cone of fire. Each creature in that area
      must make a DC 11 Dexterity saving throw, taking 7 (2d6) fire damage
      on a failed save, or half as much damage on a successful one.
    Variant: Summon Mephits.
      The mephit has a 25 percent chance of summoning 1d4 mephits of its
      kind. A summoned mephit appears in an unoccupied space within 60 feet
      of its summoner, acts as an ally of its summoner, and can't summon
      other mephits. It remains for 1 minute, until it or its summoner dies,
      or until its summoner dismisses it as an action.
    """
    name = "Magma Mephit"
    description = "Small elemental, neutral evil"
    challenge_rating = 0.5
    armor_class = 11
    skills = "Stealth +3"
    senses = "Darkvision 60 ft., Passive Perception 10"
    languages = "Ignan, Terran"
    strength = Ability(8)
    dexterity = Ability(12)
    constitution = Ability(12)
    intelligence = Ability(7)
    wisdom = Ability(10)
    charisma = Ability(10)
    speed = 30
    swim_speed = 0
    fly_speed = 30
    climb_speed = 0
    hp_max = 22
    hit_dice = "5d6"
    spells = []


class Magmin(Monster):
    """
    Death Burst.
      When the magmin dies, it explodes in a burst of fire and magma. Each
      creature within 10 ft. of it must make a DC 11 Dexterity saving throw,
      taking 7 (2d6) fire damage on a failed save, or half as much damage on
      a successful one. Flammable objects that aren't being worn or carried
      in that area are ignited.
    Ignited Illumination.
      As a bonus action, the magmin can set itself ablaze or extinguish its
      flames. While ablaze, the magmin sheds bright light in a 10-foot
      radius and dim light for an additional 10 ft.
    Touch.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 7 (2d6)
      fire damage. If the target is a creature or a flammable object, it
      ignites. Until a target takes an action to douse the fire, the target
      takes 3 (1d6) fire damage at the end of each of its turns.
    """
    name = "Magmin"
    description = "Small elemental, chaotic neutral"
    challenge_rating = 0.5
    armor_class = 14
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 10"
    languages = "Ignan"
    strength = Ability(7)
    dexterity = Ability(15)
    constitution = Ability(12)
    intelligence = Ability(8)
    wisdom = Ability(11)
    charisma = Ability(10)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 9
    hit_dice = "2d6"
    spells = []


class Mammoth(Monster):
    """
    Trampling Charge.
      If the mammoth moves at least 20 ft. straight toward a creature and
      then hits it with a gore attack on the same turn, that target must
      succeed on a DC 18 Strength saving throw or be knocked prone. If the
      target is prone, the mammoth can make one stomp attack against it as a
      bonus action.
    Gore.
      Melee Weapon Attack: +10 to hit, reach 10 ft., one target. Hit: 25
      (4d8 + 7) piercing damage.
    Stomp.
      Melee Weapon Attack: +10 to hit, reach 5 ft., one prone creature. Hit:
      29 (4d10 + 7) bludgeoning damage.
    """
    name = "Mammoth"
    description = "Huge beast, unaligned"
    challenge_rating = 6
    armor_class = 13
    skills = ""
    senses = "Passive Perception 10"
    languages = ""
    strength = Ability(24)
    dexterity = Ability(9)
    constitution = Ability(21)
    intelligence = Ability(3)
    wisdom = Ability(11)
    charisma = Ability(6)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 126
    hit_dice = "11d12"
    spells = []


class Manticore(Monster):
    """
    Tail Spike Regrowth.
      The manticore has twenty-four tail spikes. Used spikes regrow when the
      manticore finishes a long rest.
    Multiattack.
      The manticore makes three attacks: one with its bite and two with its
      claws or three with its tail spikes.
    Bite.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 7 (1d8 +
      3) piercing damage.
    Claw.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 6 (1d6 +
      3) slashing damage.
    Tail Spike.
      Ranged Weapon Attack: +5 to hit, range 100/200 ft., one target. Hit: 7
      (1d8 + 3) piercing damage.
    """
    name = "Manticore"
    description = "Large monstrosity, lawful evil"
    challenge_rating = 3
    armor_class = 14
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 11"
    languages = ""
    strength = Ability(17)
    dexterity = Ability(16)
    constitution = Ability(17)
    intelligence = Ability(7)
    wisdom = Ability(12)
    charisma = Ability(8)
    speed = 30
    swim_speed = 0
    fly_speed = 50
    climb_speed = 0
    hp_max = 68
    hit_dice = "8d10"
    spells = []


class Marilith(Monster):
    """
    Magic Resistance.
      The marilith has advantage on saving throws against spells and other
      magical effects.
    Magic Weapons.
      The marilith's weapon attacks are magical.
    Reactive.
      The marilith can take one reaction on every turn in combat.
    Multiattack.
      The marilith can make seven attacks: six with its longswords and one
      with its tail.
    Longsword.
      Melee Weapon Attack: +9 to hit, reach 5 ft., one target. Hit: 13 (2d8
      + 4) slashing damage.
    Tail.
      Melee Weapon Attack: +9 to hit, reach 10 ft., one creature. Hit: 15
      (2d10 + 4) bludgeoning damage. If the target is Medium or smaller, it
      is grappled (escape DC 19). Until this grapple ends, the target is
      restrained, the marilith can automatically hit the target with its
      tail, and the marilith can't make tail attacks against other targets.
    Teleport.
      The marilith magically teleports, along with any equipment it is
      wearing or carrying, up to 120 feet to an unoccupied space it can see.
    Variant: Summon Demon.
      The demon chooses what to summon and attempts a magical summoning.
      
      A marilith has a 50 percent chance of summoning 1d6 vrocks, 1d4
      hezrous, 1d3 glabrezus, 1d2 nalfeshnees, or one marilith.
      
      A summoned demon appears in an unoccupied space within 60 feet of its
      summoner, acts as an ally of its summoner, and can't summon other
      demons. It remains for 1 minute, until it or its summoner dies, or
      until its summoner dismisses it as an action.
    """
    name = "Marilith"
    description = "Large fiend, chaotic evil"
    challenge_rating = 16
    armor_class = 18
    skills = ""
    senses = "Truesight 120 ft., Passive Perception 13"
    languages = "Abyssal, telepathy 120 ft."
    strength = Ability(18)
    dexterity = Ability(20)
    constitution = Ability(20)
    intelligence = Ability(18)
    wisdom = Ability(16)
    charisma = Ability(20)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 189
    hit_dice = "18d10"
    spells = []


class Mastiff(Monster):
    """
    Keen Hearing and Smell.
      The mastiff has advantage on Wisdom (Perception) checks that rely on
      hearing or smell.
    Bite.
      Melee Weapon Attack: +3 to hit, reach 5 ft., one target. Hit: 4 (1d6 +
      1) piercing damage. If the target is a creature, it must succeed on a
      DC 11 Strength saving throw or be knocked prone.
    """
    name = "Mastiff"
    description = "Medium beast, unaligned"
    challenge_rating = 0.125
    armor_class = 12
    skills = "Perception +3"
    senses = "Passive Perception 13"
    languages = ""
    strength = Ability(13)
    dexterity = Ability(14)
    constitution = Ability(12)
    intelligence = Ability(3)
    wisdom = Ability(12)
    charisma = Ability(7)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 5
    hit_dice = "1d8"
    spells = []


class Medusa(Monster):
    """
    Petrifying Gaze.
      When a creature that can see the medusa's eyes starts its turn within
      30 ft. of the medusa, the medusa can force it to make a DC 14
      Constitution saving throw if the medusa isn't incapacitated and can
      see the creature. If the saving throw fails by 5 or more, the creature
      is instantly petrified. Otherwise, a creature that fails the save
      begins to turn to stone and is restrained. The restrained creature
      must repeat the saving throw at the end of its next turn, becoming
      petrified on a failure or ending the effect on a success. The
      petrification lasts until the creature is freed by the greater
      restoration spell or other magic.
      
      Unless surprised, a creature can avert its eyes to avoid the saving
      throw at the start of its turn. If the creature does so, it can't see
      the medusa until the start of its next turn, when it can avert its
      eyes again. If the creature looks at the medusa in the meantime, it
      must immediately make the save.
      
      If the medusa sees itself reflected on a polished surface within 30
      ft. of it and in an area of bright light, the medusa is, due to its
      curse, affected by its own gaze.
    Multiattack.
      The medusa makes either three melee attacks--one with its snake hair
      and two with its shortsword--or two ranged attacks with its longbow.
    Snake Hair.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one creature. Hit: 4 (1d4
      + 2) piercing damage plus 14 (4d6) poison damage.
    Shortsword.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 5 (1d6 +
      2) piercing damage.
    Longbow.
      Ranged Weapon Attack: +5 to hit, range 150/600 ft., one target. Hit: 6
      (1d8 + 2) piercing damage plus 7 (2d6) poison damage.
    """
    name = "Medusa"
    description = "Medium monstrosity, lawful evil"
    challenge_rating = 6
    armor_class = 15
    skills = "Deception +5, Insight +4, Perception +4, Stealth +5"
    senses = "Darkvision 60 ft., Passive Perception 14"
    languages = "Common"
    strength = Ability(10)
    dexterity = Ability(15)
    constitution = Ability(16)
    intelligence = Ability(12)
    wisdom = Ability(13)
    charisma = Ability(15)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 127
    hit_dice = "17d8"
    spells = []


class Merfolk(Monster):
    """
    Amphibious.
      The merfolk can breathe air and water.
    Spear.
      Melee or Ranged Weapon Attack: +2 to hit, reach 5 ft. or range 20/60
      ft., one target. Hit: 3 (1d6) piercing damage, or 4 (1d8) piercing
      damage if used with two hands to make a melee attack.
    """
    name = "Merfolk"
    description = "Medium humanoid, neutral"
    challenge_rating = 0.125
    armor_class = 11
    skills = "Perception +2"
    senses = "Passive Perception 12"
    languages = "Aquan, Common"
    strength = Ability(10)
    dexterity = Ability(13)
    constitution = Ability(12)
    intelligence = Ability(11)
    wisdom = Ability(11)
    charisma = Ability(12)
    speed = 10
    swim_speed = 40
    fly_speed = 0
    climb_speed = 0
    hp_max = 11
    hit_dice = "2d8"
    spells = []


class Merrow(Monster):
    """
    Amphibious.
      The merrow can breathe air and water.
    Multiattack.
      The merrow makes two attacks: one with its bite and one with its claws
      or harpoon.
    Bite.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 8 (1d8 +
      4) piercing damage.
    Claws.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 9 (2d4 +
      4) slashing damage.
    Harpoon.
      Melee or Ranged Weapon Attack: +6 to hit, reach 5 ft. or range 20/60
      ft., one target. Hit: 11 (2d6 + 4) piercing damage. If the target is a
      Huge or smaller creature, it must succeed on a Strength contest
      against the merrow or be pulled up to 20 feet toward the merrow.
    """
    name = "Merrow"
    description = "Large monstrosity, chaotic evil"
    challenge_rating = 2
    armor_class = 13
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 10"
    languages = "Abyssal, Aquan"
    strength = Ability(18)
    dexterity = Ability(10)
    constitution = Ability(15)
    intelligence = Ability(8)
    wisdom = Ability(10)
    charisma = Ability(9)
    speed = 10
    swim_speed = 40
    fly_speed = 0
    climb_speed = 0
    hp_max = 45
    hit_dice = "6d10"
    spells = []


class Mimic(Monster):
    """
    Shapechanger.
      The mimic can use its action to polymorph into an object or back into
      its true, amorphous form. Its statistics are the same in each form.
      Any equipment it is wearing or carrying isn 't transformed. It reverts
      to its true form if it dies.
    Adhesive (Object Form Only).
      The mimic adheres to anything that touches it. A Huge or smaller
      creature adhered to the mimic is also grappled by it (escape DC 13).
      Ability checks made to escape this grapple have disadvantage.
    False Appearance (Object Form Only).
      While the mimic remains motionless, it is indistinguishable from an
      ordinary object.
    Grappler.
      The mimic has advantage on attack rolls against any creature grappled
      by it.
    Pseudopod.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 7 (1d8 +
      3) bludgeoning damage. If the mimic is in object form, the target is
      subjected to its Adhesive trait.
    Bite.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 7 (1d8 +
      3) piercing damage plus 4 (1d8) acid damage.
    """
    name = "Mimic"
    description = "Medium monstrosity, neutral"
    challenge_rating = 2
    armor_class = 12
    skills = "Stealth +5"
    senses = "Darkvision 60 ft., Passive Perception 11"
    languages = ""
    strength = Ability(17)
    dexterity = Ability(12)
    constitution = Ability(15)
    intelligence = Ability(5)
    wisdom = Ability(13)
    charisma = Ability(8)
    speed = 15
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 58
    hit_dice = "9d8"
    spells = []


class Minotaur(Monster):
    """
    Charge.
      If the minotaur moves at least 10 ft. straight toward a target and
      then hits it with a gore attack on the same turn, the target takes an
      extra 9 (2d8) piercing damage. If the target is a creature, it must
      succeed on a DC 14 Strength saving throw or be pushed up to 10 ft.
      away and knocked prone.
    Labyrinthine Recall.
      The minotaur can perfectly recall any path it has traveled.
    Reckless.
      At the start of its turn, the minotaur can gain advantage on all melee
      weapon attack rolls it makes during that turn, but attack rolls
      against it have advantage until the start of its next turn.
    Greataxe.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 17 (2d12
      + 4) slashing damage.
    Gore.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 13 (2d8
      + 4) piercing damage.
    """
    name = "Minotaur"
    description = "Large monstrosity, chaotic evil"
    challenge_rating = 3
    armor_class = 14
    skills = "Perception +7"
    senses = "Darkvision 60 ft., Passive Perception 17"
    languages = "Abyssal"
    strength = Ability(18)
    dexterity = Ability(11)
    constitution = Ability(16)
    intelligence = Ability(6)
    wisdom = Ability(16)
    charisma = Ability(9)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 76
    hit_dice = "9d10"
    spells = []


class MinotaurSkeleton(Monster):
    """
    Charge.
      If the skeleton moves at least 10 feet straight toward a target and
      then hits it with a gore attack on the same turn, the target takes an
      extra 9 (2d8) piercing damage. If the target is a creature, it must
      succeed on a DC 14 Strength saving throw or be pushed up to 10 feet
      away and knocked prone.
    Greataxe.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 17 (2d12
      + 4) slashing damage.
    Gore.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 13 (2d8
      + 4) piercing damage.
    """
    name = "Minotaur Skeleton"
    description = "Large undead, lawful evil"
    challenge_rating = 2
    armor_class = 12
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 9"
    languages = "understands Abyssal but can't speak"
    strength = Ability(18)
    dexterity = Ability(11)
    constitution = Ability(15)
    intelligence = Ability(6)
    wisdom = Ability(8)
    charisma = Ability(5)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 67
    hit_dice = "9d10"
    spells = []


class Mule(Monster):
    """
    Beast of Burden.
      The mule is considered to be a Large animal for the purpose of
      determining its carrying capacity.
    Sure-Footed.
      The mule has advantage on Strength and Dexterity saving throws made
      against effects that would knock it prone.
    Hooves.
      Melee Weapon Attack: +2 to hit, reach 5 ft., one target. Hit: 4 (1d4 +
      2) bludgeoning damage.
    """
    name = "Mule"
    description = "Medium beast, unaligned"
    challenge_rating = 0.125
    armor_class = 10
    skills = ""
    senses = "Passive Perception 10"
    languages = ""
    strength = Ability(14)
    dexterity = Ability(10)
    constitution = Ability(13)
    intelligence = Ability(2)
    wisdom = Ability(10)
    charisma = Ability(5)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 11
    hit_dice = "2d8"
    spells = []


class Mummy(Monster):
    """
    Multiattack.
      The mummy can use its Dreadful Glare and makes one attack with its
      rotting fist.
    Rotting Fist.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 10 (2d6
      + 3) bludgeoning damage plus 10 (3d6) necrotic damage. If the target
      is a creature, it must succeed on a DC 12 Constitution saving throw or
      be cursed with mummy rot. The cursed target can't regain hit points,
      and its hit point maximum decreases by 10 (3d6) for every 24 hours
      that elapse. If the curse reduces the target's hit point maximum to 0,
      the target dies, and its body turns to dust. The curse lasts until
      removed by the remove curse spell or other magic.
    Dreadful Glare.
      The mummy targets one creature it can see within 60 ft. of it. If the
      target can see the mummy, it must succeed on a DC 11 Wisdom saving
      throw against this magic or become frightened until the end of the
      mummy's next turn. If the target fails the saving throw by 5 or more,
      it is also paralyzed for the same duration. A target that succeeds on
      the saving throw is immune to the Dreadful Glare of all mummies (but
      not mummy lords) for the next 24 hours.
    """
    name = "Mummy"
    description = "Medium undead, lawful evil"
    challenge_rating = 3
    armor_class = 11
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 10"
    languages = "the languages it knew in life"
    strength = Ability(16)
    dexterity = Ability(8)
    constitution = Ability(15)
    intelligence = Ability(6)
    wisdom = Ability(10)
    charisma = Ability(12)
    speed = 20
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 58
    hit_dice = "9d8"
    spells = []


class MummyLord(Monster):
    """
    Magic Resistance.
      The mummy lord has advantage on saving throws against spells and other
      magical effects.
    Rejuvenation.
      A destroyed mummy lord gains a new body in 24 hours if its heart is
      intact, regaining all its hit points and becoming active again. The
      new body appears within 5 feet of the mummy lord's heart.
    Spellcasting.
      The mummy lord is a 10th-level spellcaster. Its spellcasting ability
      is Wisdom (spell save DC 17, +9 to hit with spell attacks). The mummy
      lord has the following cleric spells prepared:
      
  
      
      - Cantrips (at will): sacred flame, thaumaturgy
      
      - 1st level (4 slots): command, guiding bolt, shield of faith
      
      - 2nd level (3 slots): hold person, silence, spiritual weapon
      
      - 3rd level (3 slots): animate dead, dispel magic
      
      - 4th level (3 slots): divination, guardian of faith
      
      - 5th level (2 slots): contagion, insect plague
      
      - 6th level (1 slot): harm
    Multiattack.
      The mummy can use its Dreadful Glare and makes one attack with its
      rotting fist.
    Rotting Fist.
      Melee Weapon Attack: +9 to hit, reach 5 ft., one target. Hit: 14 (3d6
      + 4) bludgeoning damage plus 21 (6d6) necrotic damage. If the target
      is a creature, it must succeed on a DC 16 Constitution saving throw or
      be cursed with mummy rot. The cursed target can't regain hit points,
      and its hit point maximum decreases by 10 (3d6) for every 24 hours
      that elapse. If the curse reduces the target's hit point maximum to 0,
      the target dies, and its body turns to dust. The curse lasts until
      removed by the remove curse spell or other magic.
    Dreadful Glare.
      The mummy lord targets one creature it can see within 60 feet of it.
      If the target can see the mummy lord, it must succeed on a DC 16
      Wisdom saving throw against this magic or become frightened until the
      end of the mummy's next turn. If the target fails the saving throw by
      5 or more, it is also paralyzed for the same duration. A target that
      succeeds on the saving throw is immune to the Dreadful Glare of all
      mummies and mummy lords for the next 24 hours.
    """
    name = "Mummy Lord"
    description = "Medium undead, lawful evil"
    challenge_rating = 15
    armor_class = 17
    skills = "History +5, Religion +5"
    senses = "Darkvision 60 ft., Passive Perception 14"
    languages = "the languages it knew in life"
    strength = Ability(18)
    dexterity = Ability(10)
    constitution = Ability(17)
    intelligence = Ability(11)
    wisdom = Ability(18)
    charisma = Ability(16)
    speed = 20
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 97
    hit_dice = "13d8"
    spells = ["sacred flame", "thaumaturgy", "command", "guiding bolt", "shield of faith", "hold person", "silence", "spiritual weapon", "animate dead", "dispel magic", "divination", "guardian of faith", "contagion", "insect plague", "harm"]
